from tko.play.keys import GuiActions, GuiKeys
from tko.game.xp import XP
from tko.settings.settings import Settings
from tko.util.text import Text
from tko.util.symbols import symbols
import os

from tko.play.fmt import Fmt
from tko.play.frame import Frame
from tko.play.border import Border
from tko.play.search import Search
from tko.play.language_setter import LanguageSetter
from tko.play.images import opening
from tko.play.floating import Floating
from tko.play.floating_manager import FloatingManager
from tko.play.flags import Flags, FlagsMan
from tko.play.tasktree import TaskTree, TaskAction

from tko.play.keys import GuiKeys
from tko.game.quest import Quest
from tko.game.task import Task
from tko.game.cluster import Cluster
from tko.play.task_graph import TaskGraph
from tko.play.week_graph import WeekGraph

class Gui:

    def __init__(self, tree: TaskTree, flagsman: FlagsMan, fman: FloatingManager):
        self.rep = tree.rep
        self.game = tree.game
        self.tree = tree
        self.flagsman = flagsman
        self.fman = fman
        self.settings = tree.settings
        self.search = Search(tree=self.tree, fman=self.fman)
        self.style: Border = Border(self.settings.app)
        self.language = LanguageSetter(self.settings, self.rep, self.flagsman, self.fman)
        self.colors = self.settings.colors
        self.need_update = False

        self.app = Settings().app

    def set_need_update(self):
        self.need_update = True

    def get_help_fixed(self):
        color = "W"
        try:
            selected = self.tree.get_selected_throw()
            if isinstance(selected, Task):
                color = "Y"
        except IndexError:
            pass
        act_color, act_text = self.get_activate_label()
        help_fixed: list[Text] = [
            Text().addf("C", f"{GuiActions.move} [{GuiKeys.up}{GuiKeys.left}{GuiKeys.down}{GuiKeys.right}]"),
            Text().addf("C", f"{GuiActions.config} [{GuiKeys.palette}]"),
            Text().addf(act_color, f"{act_text} [↲]"),
            Text().addf("G", f"{GuiActions.search} [{GuiKeys.search}]"),
            Text().addf(color, f"{GuiActions.edit} [{GuiKeys.edit}]"),
            Text().addf(color, f"{GuiActions.grade} [{GuiKeys.grade_play}]"),
        ]
        return help_fixed

    def get_activate_label(self) -> tuple[str, str]:
        output: str = GuiActions.activate
        try:
            obj = self.tree.get_selected_throw()
        except IndexError:
            return "R", " Retornar"
        if isinstance(obj, Quest):
            quest: Quest = obj
            if not Flags.admin and not quest.is_reachable():
                output = TaskAction.BLOQUEIO
            elif quest.key in self.tree.expanded:
                output = TaskAction.CONTRAIR
            else:
                output = TaskAction.EXPANDIR
            return "Y", output
        elif isinstance(obj, Cluster):
            cluster: Cluster = obj
            if cluster.key in self.tree.expanded:
                output = TaskAction.CONTRAIR
            else:
                output = TaskAction.EXPANDIR
            return "Y", output
        elif isinstance(obj, Task):
            color, output = self.tree.get_task_action(obj)
            return color, output
        return "R", " ERRO"

    def get_admin_color(self) -> str:
        if Flags.admin:
            return ""
        return "g"

    def center_header_footer(self, value: Text, frame: Frame) -> Text:
        half = value.len() // 2
        x = frame.get_x()
        _, dx = Fmt.get_size()
        color = self.get_admin_color()
        full = Text().addf(color, "─" * ((dx//2) - x - 2 - half)).add(value)
        return full

    def show_main_bar(self, frame: Frame):
        top = Text()
        alias_color = "R"
        dirname = self.rep.get_rep_dir()
        dirname = os.path.basename(dirname).upper()
        top.add(self.style.border(alias_color, dirname))
        if not Flags.admin:
            color = "W" if Flags.admin else "K"
            top.add(self.style.border(color, "GAMER"))
        top.add(self.style.border("G", self.rep.get_lang().upper()))
        full = self.center_header_footer(top, frame)
        frame.set_header(full, "<")
        if self.need_update:
            text = Text().addf("r", " TKO DESATUALIZADO!").addf("y"," Atualize com: ").addf("g", "pipx upgrade tko ")
            text = self.center_header_footer(text, frame)
            frame.set_footer(text, "<")
        
        # if Flags.flags:
        #     value = self.make_flags_bar()
        #     full = self.center_header_footer(value, frame)
        #     frame.set_footer(full, "<")
        frame.draw()

        dy, dx = frame.get_inner()
        for y, sentence in enumerate(self.tree.get_senteces(dy)):
            if sentence.len() > dx:
                sentence.trim_end(dx - 3)
                sentence.addf("r", "...")
            frame.write(y, 0, sentence)



    def show_skills_bar(self, frame_xp: Frame):
        dy, dx = frame_xp.get_inner()
        xp = XP(self.game)
        total_perc = int(
            100 * (xp.get_xp_total_obtained() / xp.get_xp_total_available())
        )
        if Flags.percent:
            text = f" XPTotal:{total_perc}%"
        else:
            text = f" XPTotal:{xp.get_xp_total_obtained()}"

        done = self.colors.main_bar_done
        todo = self.colors.main_bar_todo
        total_bar = self.style.build_bar(text, total_perc / 100, dx - 2, done, todo)
        frame_xp.set_header(Text().addf("/", "Skills"), "^", "{", "}")
        frame_xp.set_footer(Text().add(" ").add(" "), "^")
        frame_xp.draw()

        reachable_quests = [q for q in self.game.quests.values() if q.is_reachable()]
        total, obt = self.game.get_skills_resume(reachable_quests)
        elements: list[Text] = []
        for skill, value in total.items():
            if Flags.percent:
                text = f"{skill}:{int(100 * obt[skill] / value)}%"
            else:
                text = f"{skill}:{obt[skill]}/{value}"
            perc = obt[skill] / value
            done = self.colors.progress_skill_done
            todo = self.colors.progress_skill_todo
            skill_bar = self.style.build_bar(text, perc, dx - 2, done, todo)
            elements.append(skill_bar)
            
        elements.append(total_bar)

        line_breaks = dy - len(elements)
        for skill_bar in elements:
            frame_xp.print(1, skill_bar)
            if line_breaks > 0:
                line_breaks -= 1
                frame_xp.print(1, Text())


    def build_list_sentence(self, items: list[Text]) -> list[Text]:
        out: list[Text] = []
        for x in items:
            color_ini = x.data[0].fmt
            color_end = x.data[-1].fmt
            left = self.style.roundL(color_ini)
            right = self.style.roundR(color_end)
            middle = x.clone()
            if x.data[0].text == "!":
                left = self.style.sharpL(color_ini)
                right = self.style.sharpR(color_end)
                middle.data = x.data[1:]
            out.append(Text().add(left).add(middle).add(right))
        return out

    def build_bottom_array(self):
        array: list[Text] = []
        array += self.get_help_fixed()
        return self.build_list_sentence(array)

    def show_bottom_bar(self):
        lines, cols = Fmt.get_size()
        elems = self.build_bottom_array()
        line_main = Text().add(" ").join(elems) # alignment adjust
        Fmt.write(lines - 1, 0, line_main.center(cols))

    def make_xp_button(self, size: int):
        if self.search.search_mode:
            text = " Busca: " + self.tree.search_text + symbols.cursor.text
            percent = 0.0
            done = "W"
            todo = "W"
            text = text.ljust(size)
        else:
            text, percent = self.build_xp_bar()
            done = self.colors.main_bar_done
            todo = self.colors.main_bar_todo
            text = text.center(size)
        xpbar = self.style.build_bar(text, percent, len(text), done, todo)
        return xpbar

    def show_top_bar(self, frame: Frame):
        lista = [
            Text().addf("Y", f"Sair  [q]"),
            Text().addf("Y", f"Ajuda [?]"),
        ]
        help = self.build_list_sentence(lista)
        search = help[0]
        evaluate = help[1]

        pre: list[Text] = []
        pre.append(search)

        pos: list[Text] = []
        pos.append(evaluate)

        limit = frame.get_dx()
        size = limit - Text().add(" ").join(pre + pos).len() - 2
        main_label = self.make_xp_button(size)
        info = Text().add(" ").join(pre + [main_label] + pos)
        frame.write(0, 0, info.center(frame.get_dx()))

    def show_help(self):
        # def empty(value):
        #     pass
        _help: Floating = Floating(self.settings, "v>").set_text_ljust()
        self.fman.add_input(_help)

        _help.set_header_text(Text().add(" Ajuda "))
        # _help.put_text(" Movimentação ".center(dx, symbols.hbar.text))
        _help.put_sentence(Text.format("    Ajuda ").addf("r", GuiKeys.key_help).add("  Abre essa tela de ajuda")
        )

        _help.put_sentence(Text.format("  ").addf("r", "Shift + B")
                           .add("  Habilita ").addf("r", "").addf("R", "ícones").addf("r", "").add(" se seu ambiente suportar"))
        _help.put_sentence(Text() + "  " + ("g", "setas") + "      Para navegar entre os elementos")
        _help.put_sentence(Text() + "   " + ("g", f"{GuiKeys.left}{GuiKeys.down}{GuiKeys.up}{GuiKeys.right}")  + "      Para navegar entre os elementos")
        _help.put_sentence(Text() + "   " + ("g", f"{GuiKeys.left2}{GuiKeys.down2}{GuiKeys.up2}{GuiKeys.right2}")  + "      Para navegar entre os elementos")
    
        _help.put_sentence(Text() + f"   {GuiActions.config} " + ("r", f"{GuiKeys.palette}") + "  Abre o menu de ações e configurações")
        _help.put_sentence(Text() + f"   {GuiActions.github} " + ("r", f"{GuiKeys.open_url}") + "  Abre tarefa em uma aba do browser")
        _help.put_sentence(Text() + f"   {GuiActions.download} " + ("r", f"{GuiKeys.down_task}") + "  Baixa tarefa de código para seu dispositivo")
        _help.put_sentence(Text() + f"   {GuiActions.edit} " + ("r", f"{GuiKeys.edit}") + "  Abre os arquivos no editor de código")
        _help.put_sentence(Text() + f"   {GuiActions.activate} " + ("r", "↲") + "  Interage com o elemento de acordo com o contexto")
        _help.put_sentence(Text() + "             (baixar, visitar, escolher, compactar, expandir)")
        _help.put_sentence(Text() + f"  {GuiActions.grade} " + ("r", GuiKeys.grade_play) + "  Abre tela para auto avaliação")
        _help.put_sentence(Text() + f"    {GuiActions.search} " + ("r", f"{GuiKeys.search}") + "  Abre a barra de pesquisa")

        _help.put_sentence(Text())
        _help.put_sentence(Text() + "Você pode mudar o editor padrão com o comando")
        _help.put_sentence(Text() + ("g", "             tko config --editor <comando>"))


    def build_xp_bar(self) -> tuple[str, float]:
        xp = XP(self.game)
        available = xp.get_xp_total_available()
        if available > 0 and xp.get_xp_total_obtained() == available:
            text = "Você atingiu o máximo de xp!"
            percent = 100.0
        else:
            # lang = self.rep.get_lang().upper()
            level = xp.get_level()
            percent = float(xp.get_xp_level_current()) / float(xp.get_xp_level_needed())
            if Flags.percent:
                xpobt = int(100 * xp.get_xp_level_current() / xp.get_xp_level_needed())
                text = "Level:{} XP:{}%".format(level, xpobt)
            else:
                xpobt1 = xp.get_xp_level_current()
                xpobt2 = xp.get_xp_level_needed()
                text = "Level:{} XP:{}/{}".format(level, xpobt1, xpobt2)

        return text, percent

    def get_task_graph(self, task_key: str, width: int, height: int) -> tuple[bool, list[Text]]:
        tg = TaskGraph(self.settings, self.rep, task_key, width, height)
        if len(tg.collected_cov) == 1:
            return False, []
        graph = tg.get_graph()
        # for y, line in enumerate(graph):
        #     frame.write(y, x, Text().addf("g", line))
        return True, graph

    def get_week_graph(self, width: int, height: int) -> tuple[bool, list[Text]]:
        week_mode = Flags.graph.get_value() == "1"
        tg = WeekGraph(width, height, week_mode)
        if len(tg.collected) == 1:
            return False, []
        graph = tg.get_graph()
        # for y, line in enumerate(graph):
        #     frame.write(y, x, Text().addf("g", line))
        return True, graph

    def show_graphs(self, frame: Frame):
        lines, cols = frame.get_inner()
        
        # now: datetime.datetime = datetime.datetime.now()
        # parrot = random_get(opening, str(now.hour))
        distance = 18
        made = False
        list_data: list[Text] = []
        try:
            selected = self.tree.get_selected_throw()
        except IndexError:
            selected = None
        is_task = isinstance(selected, Task)
        if Flags.graph.get_value() == "1" or not is_task:
            width = cols - self.tree.max_title - distance - 7
            if width < 5:
                width = 5
            height = lines - 5
            if height < 3:
                height = 3
            if isinstance(selected, Task):
                made, list_data = self.get_task_graph(selected.key, width, height)
            elif isinstance(selected, Cluster) or isinstance(selected, Quest):
                made, list_data = self.get_week_graph(width, height)
        if not made:
            list_data = [Text().add(x) for x in opening["estuda"].splitlines()]
        op_one = "Nenhum" if is_task else "h/semana"
        op_two = "Ligado" if is_task else "h/dia   "
        border = Border(self.settings.app)
        view_button = Text().add("  ").add(border.border("C", f"Mudar Visão [{GuiKeys.graph}]"))
        view_value = Flags.graph.get_value()
        view_button.add(" ").addf("M" if view_value == "0" else "Y", f" {op_one} ")
        view_button.add(" ").addf("M" if view_value == "1" else "Y", f" {op_two} ")
        frame.write(0, self.tree.max_title + distance, view_button)
        for y, line in enumerate(list_data):
            frame.write(y + 1, self.tree.max_title + distance, line)

    def show_items(self):
        border_color = self.get_admin_color()
        Fmt.clear()
        self.tree.reload_sentences()
        lines, cols = Fmt.get_size()
        main_sx = cols  # tamanho em x livre
        main_sy = lines  # size em y avaliable

        top_y = -1
        top_dy = 1  #quantas linhas o topo usa
        bottom_dy = 1 # quantas linhas o fundo usa
        mid_y = top_dy # onde o meio começa
        mid_sy = main_sy - top_dy - bottom_dy # tamanho do meio
        left_size = 25
        skills_sx = 0
        flags_sx = 0
        if Flags.skills:
            skills_sx = left_size #max(20, main_sx // 4)
        
        task_sx = main_sx - flags_sx - skills_sx

        frame_top = Frame(top_y, 0).set_size(top_dy + 2, cols)
        self.show_top_bar(frame_top)

        # frame_bottom = Frame(lines - bottom_dy - 1, -1).set_size(bottom_dy + 2, cols + 2)
        self.show_bottom_bar()
        # if task_sx > 5: 
        frame_main = Frame(mid_y, 0).set_size(mid_sy, task_sx).set_border_color(border_color)
        self.show_main_bar(frame_main)
        self.show_graphs(frame_main)

        if Flags.skills:
            frame_skills = Frame(mid_y, cols - skills_sx).set_size(mid_sy, skills_sx).set_border_color(border_color)
            self.show_skills_bar(frame_skills)



from __future__ import annotations
import enum
import hashlib


class LogAction:
    class Type(enum.Enum):
        NONE = 'NONE'
        OPEN = 'OPEN' # Open program
        QUIT = 'QUIT' # Quit program
        DOWN = 'DOWN' # Down problem
        FREE = 'FREE' # Free Compile and RUN
        FAIL = 'FAIL' # Compile error or Execution error
        TEST = 'TEST' # run test and Coverage %
        SELF = 'SELF' # {c:{coverage}, a:{approach}, s:{autonomy}} after added {clear:%d, fun:%d, easy:%d}
        PICK = 'PICK' # Enter problem
        BACK = 'BACK' # Leave problem
        PROG = 'PROG' # deprecated
        SIZE = 'SIZE' # Problem files changes "{line_count:%d}"

    def __init__(self, action_value: str, task: str = "", payload: str = ""):
        self.hash = ""
        self.timestamp = ""
        self.type_value: str = action_value
        self.task_key = task
        self.payload = payload

    def set_hash(self, hash: str):
        self.hash = hash
        return self

    def set_timestamp(self, timestamp: str):
        self.timestamp = timestamp
        return self

    def __str__(self):
        return "{" + f'time:{self.timestamp}, type:{self.type_value}, key:{self.task_key}, payload:{self.payload}' + "}"

    @staticmethod
    def generate_hash(action_data: LogAction, last_hash: str):
        length = 6
        input_str = str(action_data) + last_hash
        full_hash = hashlib.sha256(input_str.encode()).hexdigest()
        return full_hash[:length]  # Retorna os primeiros 'comprimento' caracteres do hash


import os
import argparse
import shutil
from tko.util.decoder import Decoder
from typing import Any

class Mark:
    def __init__(self, marker: str, indent: int):
        self.marker: str = marker
        self.indent: int = indent

    # @override
    def __str__(self):
        return f"{self.marker}:{self.indent}"

class Mode:
    ADD = "ADD!"
    COM = "COM!"
    ACT = "ACT!"
    DEL = "DEL!"
    opts = [ADD, COM, ACT, DEL]

def get_comment(filename: str) -> str:
    com = "//"
    if filename.endswith(".py"):
        com = "#"
    elif filename.endswith(".hs"):
        com = "--"
    elif filename.endswith(".puml"):
        com = "'"
    return com

class Filter:
    def __init__(self, filename: str):
        self.filename = filename
        self.stack = [Mark(Mode.ADD, 0)]
        self.com = get_comment(filename)
        self.tab_char = "\t" if filename.endswith(".go") else " "

    def get_marker(self) -> str:
        return self.stack[-1].marker

    def get_indent(self) -> int:
        return self.stack[-1].indent

    def outside_scope(self, line: str):
        stripped = line.strip()
        left_spaces = len(line) - len(line.lstrip())
        return stripped != "" and left_spaces < self.get_indent()

    def has_single_mode_cmd(self, line: str) -> bool:
        stripped = line.strip()
        for marker in Mode.opts:
            if stripped == self.com + " " + marker:
                return True
        return False

    def change_mode(self, line: str):
        with_left = line.rstrip()
        marker = with_left.lstrip()[len(self.com) + 1:]
        len_spaces = len(with_left) - len(self.com + marker + " ")
        while len(self.stack) > 0 and self.stack[-1].indent >= len_spaces:
            self.stack.pop()
        self.stack.append(Mark(marker, len_spaces))

    def search_temp_mode(self, line: str) -> tuple[str, int, str]:
        for marker in Mode.opts:
            if line.rstrip().endswith(self.com + " " + marker):
                count: int = 0
                for i in range(len(line)):
                    if line[i] == " ":
                        count += 1
                    else:
                        break
                return marker, count, line[:-len(self.com + marker + " ")].rstrip()
        return "---", 0, line

    def __process(self, content: str) -> str:
        lines = content.splitlines()
        output: list[str] = []
        for line in lines:
            while self.outside_scope(line):
                self.stack.pop()
            if self.has_single_mode_cmd(line):
                self.change_mode(line)
                continue
            marker: str = self.get_marker()
            indent: int = self.get_indent()
            temp_marker, temp_indent, line = self.search_temp_mode(line)
            if temp_marker != "---":
                marker = temp_marker
                indent = temp_indent

            if marker == Mode.DEL:
                continue
            elif marker == Mode.ADD:
                output.append(line)
            elif marker == Mode.ACT:
                prefix = self.tab_char * indent + self.com + " "
                if not line.startswith(prefix):
                    prefix = prefix[:-1]
                line = line.replace(prefix, self.tab_char * indent, 1)
                output.append(line)
            elif marker == Mode.COM:
                line = self.tab_char * indent + self.com + " " + line[indent:]
                output.append(line)

        return "\n".join(output) + "\n"
    
    def process(self, content: str) -> str:
        return self.__process(content)

def clean_com(target: str, content: str) -> str:
    com = get_comment(target)
    lines = content.splitlines()
    output = [line for line in lines if not line.lstrip().startswith(com)]
    return "\n".join(output)

class DeepFilter:
    extensions = [".md", ".c", ".cpp", ".h", ".hpp", ".py", ".java", ".js", ".ts", ".hs", ".txt", ".go", ".json", ".mod", ".puml", ".sh", ".sql", ".yaml", ".exec", ".hide", ".tio"]

    def __init__(self):
        self.cheat_mode = False
        self.quiet_mode = False
        self.indent = ""
    
    def print(self, *args: str, **kwargs: Any):
        if not self.quiet_mode:
            print(self.indent, end="")
            print(*args, **kwargs)

    def set_indent(self, prefix: int):
        self.indent = prefix * " "
        return self

    def set_quiet(self, value: bool):
        self.quiet_mode = (value == True)
        return self
    
    def set_cheat(self, value: bool):
        self.cheat_mode = (value == True)
        return self

    def copy(self, source: str, destiny: str, deep: int):
        if deep == 0:
            return
        
        if os.path.isdir(source):
            chain = source.split(os.sep)
            if len(chain) > 1 and chain[-1].startswith("."):
                return
            if not os.path.isdir(destiny):
                os.makedirs(destiny)
            for file in sorted(os.listdir(source)):
                self.copy(os.path.join(source, file), os.path.join(destiny, file), deep - 1)
            return
        
        filename = os.path.basename(source)
        folder = os.path.dirname(source)
        deny_list = os.path.join(folder, ".deny")
        if os.path.isfile(deny_list):
            with open(deny_list) as f:
                deny = [x.strip() for x in f.read().splitlines()]
                if filename in deny:
                    print("(disabled):", destiny)
                    return

        if not any([filename.endswith(ext) for ext in self.extensions]):
            # print("(skipped ): ", filename)
            return

        content = Decoder.load(source)

        processed = Filter(filename).process(content)

        if self.cheat_mode:
            if processed != content:
                cleaned = clean_com(source, content)
                Decoder.save(destiny, cleaned)
        elif processed != "" and processed != "\n":
            Decoder.save(destiny, processed)

        line = ""
        if self.cheat_mode:
            if processed != content:
                line += "(cleaned ): "
            else:
                line += "(disabled): "
        else:
            if processed == "" or processed == "\n":
                line += "(disabled): "
            elif processed != content:
                line += "(filtered): "
            else:
                line += "(        ): "
        line += destiny

        self.print(line)

class CodeFilter:
    @staticmethod
    def open_file(path: str): 
        if os.path.isfile(path):
            file_content = Decoder.load(path)
            return True, file_content
        print("Warning: File", path, "not found")
        return False, "" 

    @staticmethod
    def cf_recursive(target_dir: str, output_dir: str, force: bool, cheat: bool = False, quiet: bool = False, indent: int = 0):
        if not os.path.isdir(target_dir):
            print("Error: target must be a folder in recursive mode")
            exit()
        if os.path.exists(output_dir):
            if not force:
                print("Error: output folder already exists")
                exit()
            else:
                # recursive delete all folder content without deleting the folder itself
                for file in os.listdir(output_dir):
                    path = os.path.join(output_dir, file)
                    if os.path.isdir(path):
                        shutil.rmtree(path, ignore_errors=True)
                    else:
                        os.remove(path)

        deep_filter = DeepFilter().set_cheat(cheat).set_quiet(quiet).set_indent(indent)
        deep_filter.copy(target_dir, output_dir, 10)

    @staticmethod
    def cf_single_file(target: str, output: str, update: bool, cheat: bool):
        file = target
        success, content = CodeFilter.open_file(file)
        if success:
            if cheat:
                content = clean_com(file, content)
            else:
                content = Filter(file).process(content)

            if output:
                if os.path.isfile(output):
                    old = Decoder.load(output)
                    if old != content:
                        Decoder.save(output, content)
                else:
                    Decoder.save(output, content)
            elif update:
                Decoder.save(file, content)
            else:
                print(content)

def filter_main(args: argparse.Namespace):
    if args.cheat:
        args.recursive = True

    if args.recursive:
        CodeFilter.cf_recursive(args.target, args.output, force=args.force, cheat=args.cheat, quiet=args.quiet, indent=args.indent)
        exit()

    CodeFilter.cf_single_file(args.target, args.output, args.update, args.cheat)



from __future__ import annotations
from tko.game.task import Task
from tko.util.text import Text
from tko.util.get_md_link import get_md_link
from tko.game.tree_item import TreeItem

# from typing import override

def startswith(text: str, prefix: str) -> bool: 
    if len(prefix) > len(text):
        return False
    return text[:len(prefix)] == prefix

class Quest(TreeItem):
    def __init__(self, title: str = "", key: str = ""):
        super().__init__()
        self.key: str = key
        self.title: str = title
        self.line_number: int = 0
        self.line: str = ""
        self.__tasks: list[Task] = []
        self.skills: dict[str, int] = {}  # s:skill
        self.requires: list[str] = []  # r:quest_key
        self.languages: list[str] = []  # l:language
        self.requires_ptr: list[Quest] = []
        self.opt = False  # opt
        self.prog = False  # progressive tasks
        self.qmin: None | int = None  # q:  minimo de 50 porcento da pontuação total para completar
        self.tmin: None | int = None  # t: ou ter no mínimo esse valor de todas as tarefas
        self.filename = ""
        self.cluster_key = ""
        self.__is_reachable: bool = False

    def get_full_title(self):
        output = self.title
        # if Flags.minimum:
        #     output += " " + self.get_requirement().get_str()
        # if Flags.reward:
        #     xp = ""
        #     for s, v in self.skills.items():
        #         xp += f" +{s}:{v}"
        #     output += xp
        return output


    def is_reachable(self)-> bool:
        return self.__is_reachable

    def set_reachable(self, value: bool):
        self.__is_reachable = value
        return self
    
    def update_tasks_reachable(self):
        if not self.prog:
            for t in self.__tasks:
                t.set_reachable(True)
            return
        reach = True
        for i in range(len(self.__tasks)):
            if i == 0:
                self.__tasks[i].set_reachable(True)
            else:
                if self.__tasks[i-1].get_percent() < 50:
                    reach = False
                self.__tasks[i].set_reachable(reach)

    # @override
    def __str__(self):
        line = str(self.line_number).rjust(3)
        tasks_size = str(len(self.__tasks)).rjust(2, "0")
        key = "" if self.key == self.title else self.key + " "
        output = f"{line} {tasks_size} {key}{self.title} {self.skills} {self.requires}"
        return output

    def get_resume_by_percent(self) -> Text:
        value = self.get_percent()
        return Text().addf(self.get_grade_color(), (str(value) + "%").rjust(4))
    
    def get_requirement(self) -> Text:
        if self.qmin is not None:
            return Text().addf("y", f"[{self.qmin}%]")
        if self.tmin is not None:
            return Text().addf("y", f"[t>{self.tmin - 1}]")
        return Text()

    def get_resume_by_tasks(self) -> Text:
        tmin = self.tmin if self.tmin is not None else 70
        total = len([t for t in self.__tasks if not t.opt])
        plus = len([t for t in self.__tasks if t.opt])
        count = len([t for t in self.__tasks if t.get_percent() >= tmin])
        output = f"{count}/{total}"
        if plus > 0:
            output += f"+{plus}"
        return Text().addf(self.get_grade_color(), "(" + output + ")")

    def get_grade_color(self) -> str:
        if self.not_started():
            return "m"
        if not self.is_complete():
            return "r"
        if self.get_percent() == 100:
            return "g"
        return "y"

    def is_complete(self):
        if self.qmin is not None:
            return self.get_percent() >= self.qmin
        # task complete mode
        if self.tmin is not None:
            for t in self.__tasks:
                if not t.opt and t.get_percent() < self.tmin:
                    return False
        return True

    def add_task(self, task: Task, filename: str):
        # task.qskills = self.skills

        # task.xp = 0
        # for s in task.skills:
        #     task.xp += task.skills[s]

        # for s in task.qskills:
        #     task.xp += task.qskills[s]
        
        self.__tasks.append(task)

    def get_tasks(self):
        return self.__tasks

    def get_xp(self) -> tuple[int, int]:
        total = 0
        obtained = 0
        for t in self.__tasks:
            if not t.opt:
                total += t.get_xp()   
            if t.get_percent() > 0:
                obtained += int(t.get_xp() * t.get_ratio())

        return obtained, total
        
    def get_percent(self):
        obtained, total = self.get_xp()
        if total == 0:
            return 0
        return obtained * 10 // total

    def in_progress(self):
        if self.is_complete():
            return False
        for t in self.__tasks:
            if t.get_percent() != 0:
                return True
        return False

    def not_started(self):
        if self.is_complete():
            return False
        if self.in_progress():
            return False
        return True


class QuestParser:
    quest: Quest

    def __init__(self):
        self.quest = Quest()
        self.line: str = ""
        self.line_num = 0
        self.default_qmin_requirement = 50
        self.default_task_xp = 10
        self.filename: str = ""

    def finish_quest(self) -> Quest:

        if self.quest.key == "":
            self.quest.key = get_md_link(self.quest.title)

        if len(self.quest.skills) == 0:
            self.quest.skills["xp"] = self.default_task_xp
        
        if self.quest.qmin is None and self.quest.tmin is None:
            self.quest.qmin = self.default_qmin_requirement

        return self.quest

    def match_full_pattern(self) -> bool:
        if not startswith(self.line, "### "):
            return False
        line = self.line[4:]
        
        pieces: list[str] = line.split("<!--")

        # html tags
        if len(pieces) > 1:
            middle_end: list[str] = pieces[1].split("-->")
            middle: str = middle_end[0]
            end: str = middle_end[1]
            line = pieces[0] + end # removendo raw text
            self.process_raw_tags(middle)

        self.quest.title = line
        if "[](" in line:
            pieces = line.split("[](")
            self.quest.title = pieces[0]

            del pieces[0]
            for p in pieces:
                key = p.split(")")[0]
                if key[0] == "#":
                    key = key[1:]
                self.quest.requires.append(key)
 
        return True

    def process_raw_tags(self, raw_tags: str):
        tags = [tag.strip() for tag in raw_tags.split(" ")]

        # skills
        skills = [t[1:] for t in tags if t.startswith("+")]
        if len(skills) > 0:
            self.quest.skills = {}
            for s in skills:
                k, v = s.split(":")
                self.quest.skills[k] = int(v)

        # languages
        languages = [t[2:] for t in tags if t.startswith("l:")]
        if len(languages) > 0:
            self.quest.languages = []
            for l in languages:
                self.quest.languages.append(l)

        self.quest.opt = "opt" in tags
        self.quest.prog = "prog" in tags

        # quest percent
        qmin = [t[2:] for t in tags if t.startswith("q:")]
        
        if len(qmin) > 0:
            self.quest.qmin = int(qmin[0])

        # task min value requirement
        tmin = [t[2:] for t in tags if t.startswith("t:")]
        if len(tmin) > 0:
            self.quest.tmin = int(tmin[0])
            if self.quest.tmin > 10:
                print("fail: tmin > 10")
                exit(1)


    def parse_quest(self, filename: str, line: str, line_num: int) -> None | Quest:
        self.line = line
        self.line_num = line_num
        self.filename = filename

        self.quest.line = self.line
        self.quest.line_number = self.line_num
        self.quest.cluster_key = ""

        if self.match_full_pattern():
            return self.finish_quest()
                
        return None



from tko.enums.execution_result import ExecutionResult
from tko.util.text import Text
import os

class Unit:
    def __init__(self, case: str = "", input_data: str = "", expected: str = "", grade: None | int = None, source: str = ""):
        self.source = source  # stores the source file of the unit
        self.source_pad = 0  # stores the pad to justify the source file
        self.case = case  # name
        self.case_pad = 0  # stores the pad to justify the case name
        self.inserted = input_data  # input
        self.__expected = expected  # expected output
        self.__received: None | str = None  # solver generated answer
        self.grade: None | int = grade  # None represents proportional gr, 100 represents all
        self.grade_reduction: int = 0  # if grade is None, this atribute should be filled with the right grade reduction
        self.index = 0
        self.repeated: None | int = None

        self.result: ExecutionResult = ExecutionResult.UNTESTED

    def set_expected(self, expected: str):
        self.__expected = "\n".join(expected.splitlines()) + "\n"
        return self
    
    def get_expected(self) -> str:
        return self.__expected
    
    def set_received(self, received: str):
        self.__received = "\n".join(received.splitlines()) + "\n"
        return self

    def get_received(self) -> None | str:
        return self.__received

    def str(self, pad: bool = True) -> Text:
        index = str(self.index).zfill(2)
        grade = str(self.grade_reduction).zfill(3)
        rep = "" if self.repeated is None else " [" + str(self.repeated) + "]"
        op = Text() + ExecutionResult.get_symbol(self.result) + " " + self.result.value
        source = os.path.basename(self.source)
        if pad:
            source = self.source.ljust(self.source_pad)
        case = self.case
        if pad:
            case = self.case.ljust(self.case_pad)
        return Text() + "(" + op + ")" + f"[{index}] GR:{grade} {source} ({case}){rep}"



import os

from tko.enums.execution_result import ExecutionResult
from tko.run.diff_builder_down import DiffBuilderDown
from tko.run.wdir import Wdir
from tko.enums.diff_count import DiffCount
from tko.util.param import Param
from tko.run.diff_builder_side import DiffBuilderSide
from tko.util.text import Text, Token

from tko.util.raw_terminal import RawTerminal
from tko.util.symbols import symbols

from tko.util.freerun import Free
from tko.play.tester import Tester
from tko.play.tracker import Tracker
from tko.run.unit_runner import UnitRunner
from tko.game.task import Task
from tko.play.opener import Opener
from tko.settings.settings import Settings
from tko.enums.diff_mode import DiffMode
from tko.settings.logger import Logger
from tko.feno.filter import CodeFilter
from tko.settings.repository import Repository

class tkoFilterMode:
    @staticmethod
    def deep_copy_and_change_dir():
        # path to ~/.tko_filter
        filter_path = os.path.join(os.path.expanduser("~"), ".tko_filter")

        CodeFilter.cf_recursive(".", filter_path, force=True)
        os.chdir(filter_path)

class Run:

    def __init__(self, settings: Settings, target_list: list[str], param: None | Param.Basic):
        self.settings = settings
        self.target_list: list[str] = target_list
        if param is None:
            self.param = Param.Basic()
        else:
            self.param = param
        self.wdir: Wdir = Wdir()
        self.wdir_builded = False
        self.__curses_mode: bool = False
        self.__lang = ""
        self.__rep: Repository | None = None
        self.__task: Task | None = None
        self.__track_folder: str = ""
        self.__opener: Opener | None = None
        self.__run_without_ask: bool = True
        self.__eval_mode: bool = False

    def set_eval_mode(self):
        self.__eval_mode = True
        return self

    def set_curses(self, value:bool=True):
        self.__curses_mode = value
        return self
   
    def set_lang(self, lang:str):
        self.__lang = lang
        return self
    
    def set_opener(self, opener: Opener):
        self.__opener = opener
        return self

    def set_run_without_ask(self, value:bool):
        self.__run_without_ask = value
        return self

    def set_task(self, rep: Repository, task: Task):
        self.__rep = rep
        self.__task = task
        return self
    
    def get_task(self) -> Task:
        if self.__task is None:
            raise Warning("Task não definida")
        return self.__task

    # return the percentage of correct cases
    def execute(self) -> int:
        if len(self.target_list) == 0:
            self.__try_load_rep(".")
            self.__try_load_task(".")

        elif len(self.target_list) == 1 and os.path.isdir(self.target_list[0]):
            self.__try_load_rep(self.target_list[0])
            self.__try_load_task(self.target_list[0])

        if not self.wdir_builded:
            self.build_wdir()

        if self.__missing_target():
            return 0
        
        if self.__list_mode():
            return 0
        
        if self.wdir.has_solver():
            if self.__rep is None:
                solver_path = self.wdir.get_solver().args_list[0]
                dirname = os.path.dirname(os.path.abspath(solver_path))
                self.__try_load_rep(dirname)
                self.__try_load_task(dirname)

            if self.__task is None:
                self.__fill_task()

            if self.__rep is not None:

                logger = Logger.get_instance()
                history_file = self.__rep.get_history_file()
                logger.set_log_files(history_file, self.__rep.get_track_folder())

        if self.__free_run():
            return 0
        percent = self.__show_diff()
        return percent
    
    def __try_load_rep(self, dirname: str) -> bool:
        repo_path = Repository.rec_search_for_repo(dirname)
        if repo_path == "":
            return False
        rep = Repository(repo_path).load_config().load_game()
        self.__rep = rep
        if rep.get_lang() != "":
            self.__lang = rep.get_lang()
        return True

    def __try_load_task(self, dirname: str) -> bool:
        if self.__rep is None:
            return False
        rep: Repository = self.__rep
        task_key = rep.get_key_from_task_folder(dirname)
        if task_key == "":
            return False
        task = rep.game.tasks.get(task_key)
        if task is None:
            return False
        self.__task = task
        self.__track_folder = rep.get_track_task_folder(task_key)
        return True

    def __remove_duplicates(self):
        # remove duplicates in target list keeping the order
        self.target_list = list(dict.fromkeys(self.target_list))

    def __change_targets_to_filter_mode(self):
        if self.param.filter:
            old_dir = os.getcwd()

            print(Text.format(" Entrando no modo de filtragem ").center(RawTerminal.get_terminal_size(), Token("═")))
            tkoFilterMode.deep_copy_and_change_dir()  
            # search for target outside . dir and redirect target
            new_target_list: list[str] = []
            for target in self.target_list:
                if ".." in target:
                    new_target_list.append(os.path.normpath(os.path.join(old_dir, target)))
                elif os.path.exists(target):
                    new_target_list.append(target)
            self.target_list = new_target_list

    def __print_top_line(self):
        print(Text().add(symbols.opening).add(self.wdir.resume()), end="")
        print(" [", end="")
        first = True
        for unit in self.wdir.get_unit_list():
            if first:
                first = False
            else:
                print(" ", end="")
            solver = self.wdir.get_solver()
            unit.result = UnitRunner.run_unit(solver, unit)
            print(Text() + ExecutionResult.get_symbol(unit.result), end="")
        print("]", end="")
        print(f" {self.get_percent()}%", end="", flush=True)
        if self.__eval_mode:
            if self.__rep is not None:
                logger = Logger.get_instance()
                entries = logger.tasks.key_actions.get(self.get_task().key, {})
                if entries:
                    elapsed = entries[-1].elapsed.total_seconds() // 60
                    lines = entries[-1].lines
                    attempts = entries[-1].attempts
                    print(f"{{minutos:{elapsed:.0f}, linhas:{lines}, tentativas:{attempts}}}", end="", flush=True)
        print()
        
    def __print_diff(self):
        if self.param.diff_count == DiffCount.QUIET or self.__eval_mode:
            return
        
        if self.wdir.get_solver().has_compile_error():
            exec, _ = self.wdir.get_solver().get_executable()
            print(exec.get_error_msg())
            return
        
        results = [unit.result for unit in self.wdir.get_unit_list()]
        if ExecutionResult.EXECUTION_ERROR not in results and ExecutionResult.WRONG_OUTPUT not in results:
            return
        
        if not self.param.compact:
            for elem in self.wdir.unit_list_resume():
                print(elem)

        
        if self.param.diff_count == DiffCount.FIRST:
            # printing only the first wrong case
            wrong = [unit for unit in self.wdir.get_unit_list() if unit.result != ExecutionResult.SUCCESS][0]
            if self.param.diff_mode == DiffMode.DOWN:
                ud_diff_builder = DiffBuilderDown(RawTerminal.get_terminal_size(), wrong).to_insert_header()
                for line in ud_diff_builder.build_diff():
                    print(line)
            else:
                ss_diff_builder = DiffBuilderSide(RawTerminal.get_terminal_size(), wrong).to_insert_header(True)
                for line in ss_diff_builder.build_diff():
                    print(line)
            return

        if self.param.diff_count == DiffCount.ALL:
            for unit in self.wdir.get_unit_list():
                if unit.result != ExecutionResult.SUCCESS:
                    if self.param.diff_mode == DiffMode.DOWN:
                        ud_diff_builder = DiffBuilderDown(RawTerminal.get_terminal_size(), unit).to_insert_header()
                        for line in ud_diff_builder.build_diff():
                            print(line)
                    else:
                        ss_diff_builder = DiffBuilderSide(RawTerminal.get_terminal_size(), unit).to_insert_header(True)
                        for line in ss_diff_builder.build_diff():
                            print(line)

    def build_wdir(self):
        self.wdir_builded = True
        self.__remove_duplicates()
        self.__change_targets_to_filter_mode()
        try:
            self.wdir = Wdir().set_curses(self.__curses_mode).set_lang(self.__lang).set_target_list(self.target_list).build().filter(self.param)
        except FileNotFoundError as e:
            if self.wdir.has_solver():
                exec, _ = self.wdir.get_solver().get_executable()
                exec.set_compile_error(str(e))
        return self.wdir

    def __missing_target(self) -> bool:
        if not self.wdir.has_solver() and not self.wdir.has_tests():
            if not self.__curses_mode:
                print(Text().addf("", "fail: ") + "Nenhum arquivo de código ou de teste encontrado.")
            return True
        return False
    
    def __list_mode(self) -> bool:
        # list mode
        if not self.wdir.has_solver() and self.wdir.has_tests():
            print(Text.format("Nenhum arquivo de código encontrado. Listando casos de teste.").center(RawTerminal.get_terminal_size(), Token("╌")), flush=True)
            print(self.wdir.resume())
            for line in self.wdir.unit_list_resume():
                print(line)
            return True
        return False

    def __free_run(self) -> bool:
        if self.wdir.has_solver() and (not self.wdir.has_tests()) and not self.__curses_mode:
            if self.__task is not None:
                Logger.get_instance().record_freerun(self.get_task().key)
            Free.free_run(self.wdir.get_solver(), show_compilling=False, to_clear=False, wait_input=False)
            return True
        return False

    def __create_opener_for_wdir(self) -> Opener:
        opener = Opener(self.settings)
        folders: list[str] = []
        targets = ["."]
        if self.target_list:
            targets = self.target_list
        for f in targets:
            if os.path.isdir(f) and f not in folders:
                folders.append(f)
            else:
                folder = os.path.dirname(os.path.abspath(f))
                if folder not in folders:
                    folders.append(folder)
        opener.set_target(folders)
        if self.wdir.get_solver().args_list:
            solver_zero = self.wdir.get_solver().args_list[0]
            lang = solver_zero.split(".")[-1]
            opener.set_language(lang)
        return opener

    def __fill_task(self):
        task = Task()
        sources = self.wdir.get_source_list()
        solver = self.wdir.get_solver()
        if len(sources) > 0:
            task.folder = os.path.abspath(sources[0])
        elif solver.args_list:
            task.folder = os.path.abspath(self.wdir.get_solver().args_list[0])
        else:
            task.folder = os.path.abspath(os.getcwd())

        if os.path.isfile(task.folder):
            task.folder = os.path.dirname(task.folder)


        task.key = os.path.basename(task.folder)
        self.__task = task
        self.__track_folder = self.__rep.get_track_task_folder(task.key) if self.__rep else ""

    def __run_diff_on_curses(self):
        cdiff = Tester(self.settings, self.__rep, self.wdir, self.get_task())
        if self.__opener is not None:
            cdiff.set_opener(self.__opener)
        else:
            cdiff.set_opener(self.__create_opener_for_wdir())
        cdiff.set_autorun(self.__run_without_ask)
        cdiff.run()

    def get_percent(self) -> int:
        correct = [unit for unit in self.wdir.get_unit_list() if unit.result == ExecutionResult.SUCCESS]
        percent = (len(correct) * 100) // len(self.wdir.get_unit_list())
        return percent

    def __run_diff_on_raw_term(self) -> int:
        print(Text.format(" Testando o código com os casos de teste ").center(RawTerminal.get_terminal_size(), Token("═")))
        self.__print_top_line()
        self.__print_diff()
        percent = self.get_percent()
        if self.__task is None or self.__track_folder == "" or self.__eval_mode:
            return percent

        Logger.get_instance().record_test_result(self.get_task().key, percent)
        tracker = Tracker()
        tracker.set_folder(self.__track_folder)
        tracker.set_files(self.wdir.get_solver().args_list)
        tracker.set_percentage(percent)
        has_changes, total_lines = tracker.store()
        if has_changes:
            Logger.get_instance().record_file_alteration(self.get_task().key, total_lines)
        return percent

    def __show_diff(self) -> int:
        if self.__task is None:
            self.__fill_task()
        if self.__curses_mode:
            self.__run_diff_on_curses()
        else:
            return self.__run_diff_on_raw_term()

        return 0



from tko.settings.settings import Settings
from tko.settings.repository import Repository
from tko.settings.logger import Logger
from uniplot import plot_to_string # type: ignore
from tko.util.text import Text

class TaskGraph:
    def __init__(self, settings: Settings, rep: Repository, task_key: str, width: int, height: int):
        self.settings = settings
        self.rep = rep
        self.task_key = task_key
        self.width = width
        self.height = height
        self.collected_cov: list[float] = []
        self.collected_elapsed: list[float] = []
        self.total_elapsed: float = 0.0
        self.max_lines: int = 0
        self.eixo: list[float] = []
        self.logger = Logger.get_instance()

    def __collect(self):
        actions = self.logger.tasks.key_actions.get(self.task_key, [])
        types = ["TEST", "PROG", "SELF", "FAIL"]
        filtered = [ad for ad in actions if ad.type in types]
        collected_cov: list[float] = [0]
        collected_elapsed: list[float] = [0]
        collected_lines: list[int] = [0]
        max_lines = 0
        last = 0
        eixo: list[float] = [0]
        count = 1
        for ad in filtered:
            if ad.coverage == -1 or ad.coverage == 0:
                collected_cov.append(last)
            else:
                last = ad.coverage
                collected_cov.append(last)
            collected_elapsed.append(ad.elapsed.total_seconds() / 60)
            collected_lines.append(ad.lines)
            # if self.minutes_mode:
            #     eixo.append(ad.elapsed.total_seconds() / 60)
            # else:
            eixo.append(count)
            count += 1
        self.total_elapsed = collected_elapsed[-1]
        self.max_lines = max(collected_lines)
        if collected_elapsed[-1] != 0:
            for i in range(len(collected_elapsed)):
                collected_elapsed[i] = collected_elapsed[i] / self.total_elapsed * 100
        if self.max_lines != 0:
            for i in range(len(collected_lines)):
                collected_lines[i] = collected_lines[i] // self.max_lines * 100
        self.collected_cov = collected_cov
        self.collected_elapsed = collected_elapsed
        self.collected_lines = collected_lines
        self.eixo = eixo
        # self.eixo = list(range(len(collected)))

    def get_graph(self) -> list[Text]:
        self.__collect()
        lines: list[Text] = []
        title = Text.format(" {C}", f" @{self.task_key} ")
        title += Text.format(" {G}", f" Tempo: {self.total_elapsed:.0f} min ")
        title += Text.format(" {M}", f" Total: {self.collected_cov[-1]:.0f} % ")
        title += Text.format(" {B}", f" Linhas: {self.max_lines:.0f} ")
        # if len(self.collected_elapsed) > 1:
        result = plot_to_string(xs=[self.eixo, self.eixo, self.eixo], ys=[self.collected_elapsed, self.collected_lines, self.collected_cov], lines=True, y_min=0, width=self.width, height=self.height)

        if isinstance(result, str):
            result = result.splitlines()

        for line in result:
            lines.append(Text.decode_raw(line))
        title = title.center(self.width)
        lines = [title] + lines
        return lines
        # return []


from __future__ import annotations
from tko.util.symbols import symbols
from tko.util.text import Token


import enum


class ExecutionResult(enum.Enum):
    UNTESTED          = "não_verificado_"
    SUCCESS           = "saída_correta__"
    WRONG_OUTPUT      = "saída_incorreta"
    COMPILATION_ERROR = "erro_compilação"
    EXECUTION_ERROR   = "erro_execução__"

    @staticmethod
    def get_symbol(result: ExecutionResult) -> Token:
        if result == ExecutionResult.UNTESTED:
            return symbols.neutral
        elif result == ExecutionResult.SUCCESS:
            return symbols.success
        elif result == ExecutionResult.WRONG_OUTPUT:
            return symbols.wrong
        elif result == ExecutionResult.COMPILATION_ERROR:
            return symbols.compilation
        elif result == ExecutionResult.EXECUTION_ERROR:
            return symbols.execution
        else:
            raise ValueError("Invalid result type")

    # @override
    def __str__(self):
        return self.value


from tko.settings.log_action import LogAction
from tko.util.decoder import Decoder

import csv
import datetime
import os
from typing import Callable

class HistoryFile:

    def __init__(self, history_file: str | None = None, listeners: list[Callable[[LogAction, bool], None]] = []):
        self.history_file: str | None = history_file
        self.listeners: list[Callable[[LogAction, bool], None]] = listeners
        self.entries: list[LogAction] = HistoryFile.__load_file(history_file)
        for entry in self.entries:
            for listener in self.listeners:
                listener(entry, False)

    def get_entries(self) -> list[LogAction]:
        return self.entries

    def get_log_file(self) -> str | None:
        return self.history_file

    def __append_action_data(self, action_data: LogAction):
        self.entries.append(action_data)
        for listener in self.listeners:
            listener(action_data, True)
        log_file = self.get_log_file()
        if log_file is None:
            return
        if not os.path.exists(os.path.dirname(log_file)):
            os.makedirs(os.path.dirname(log_file))
        with open(log_file, 'a', encoding="utf-8", newline='') as file:
            writer = csv.writer(file)
            ad = action_data
            writer.writerow([ad.hash, ad.timestamp, ad.type_value, ad.task_key, ad.payload])

    def append_new_action(self, action: LogAction.Type, task_key: str = "", payload: str = ""):
        action_data = LogAction(action.value, task_key, payload)
        action_data.hash = LogAction.generate_hash(action_data, self.get_last_hash())
        action_data.timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        self.__append_action_data(action_data)

    @staticmethod
    def __row_to_action_data( row: list[str]) -> LogAction | None:
        if len(row) < 5:
            return None
        hash = row[0]
        timestamp = row[1]
        action_value = row[2]
        task = row[3]
        payload = row[4]
        action_data = LogAction(action_value, task, payload).set_hash(hash).set_timestamp(timestamp)
        return action_data

    @staticmethod
    def __load_file(history_file: str | None) -> list[LogAction]:
        if history_file is None:
            return []
        if not os.path.exists(history_file):
            return []
        encoding = Decoder.get_encoding(history_file)
        entries: list[LogAction] = []
        with open(history_file, 'r', encoding=encoding) as file:
            reader = csv.reader(file)
            rows = list(reader)
            for row in rows:
                action_data = HistoryFile.__row_to_action_data(row)
                if action_data is not None:
                    entries.append(action_data)
        return entries

    def get_last_hash(self) -> str:
        if len(self.entries) == 0:
            return ""
        return self.entries[-1].hash


class GuiActions:
    tab = "↔TAB"
    leave   = "Sair"
    edit = "Editar"
    move = "Mover"
    principal = "Principal"
    github = "Web   "
    help = "Ajuda"
    download = "Baixar"
    activate = "Ativar"
    navigate = "←↓→"
    uncheck = "Desmarcar"
    search = "Busca"
    config = "Config"
    grade = "Pontuar"
    # tester frame
    time_limit = "Tempo"
    evaluate_tester = "Avaliar "
    execute_tester  = "Rodar"
    calibrate = "Calibrar"
    all = "Todos"

class GuiKeys:
    left = "A"
    right = "D"
    down = "S"
    up = "W"

    left2 = "H"
    right2 = "L"
    down2 = "J"
    up2 = "K"
    calibrate = "C"

    palette = "c"
    images = "I"
    down_task = "b"
    activate = "\n"
    key_help = "?"
    
    # rate = "R"
    set_lang_drafts = "R"
    open_url = "u"
    key_quit = "q"
    edit= "e"
    colors = "C"
    borders = "B"
    search = "/"
    graph = "v"
    grade_play = "p"
    execute_tester = "r" # Interact in manual input
    diff = "d"
    evaluate = "\n"
    toggle_main = "\t"
    quit   = "q"
    lock = "u"
    limite  = "t"
    expand = ">"
    collapse = "<"
    mass = "Y"
    hidden = "E"


import os

class Drafts:
    ts_draft = r"""
const input = () => ""; //essa função será gerada pelo tko na execução
export {};

function main() {
    console.log("qxcode");
}
main();
"""[1:]

    c_draft = r"""
#include <stdio.h>

int main() {
    puts("qxcode");
    return 0;
}

"""[1:]

    cpp_draft = r"""
#include <iostream>

int main() {
    std::cout << "qxcode\n";
}

"""[1:]

    java_draft = r"""
public class draft {
    public static void main(String args[]) {
        System.out.println("qxcode");
    }
}

"""[1:]

    go_draft = (
        r"package main""\n"
        r'import "fmt"'"\n"
        r"func main() {""\n"
        r'    fmt.Println("qxcode")''\n'
        r"}""\n"
    )
    
    yaml_draft = r"""
build: comando para construir o executável
run: comando para rodar o programa
"""[1:]
    
    hs_draft = r"""
main :: IO ()
main = putStrLn "qxcode"
"""[1:]


    drafts = {'c': c_draft, 'cpp': cpp_draft, 'ts': ts_draft, 'java': java_draft, 'go': go_draft, 'hs': hs_draft, 'yaml': yaml_draft}

    @staticmethod
    def load_drafts_only(folder: str, lang: str, extra: list[str] = []) -> list[str]:
        folder = os.path.normpath(os.path.abspath(folder))
        draft_list: list[str] = []
        allowed = [lang] + extra
        if lang == "c":
            allowed.append("h")
        if lang == "cpp":
            allowed.append("h")
            allowed.append("hpp")
        if not os.path.isdir(folder):
            return []
        for root, _, files in os.walk(folder):
            cut_root = root[len(folder):]
            pieces = cut_root.split(os.sep)
            if any([piece.startswith(".") for piece in pieces]) or any([piece.startswith("_") for piece in pieces]):
                continue
            for file in files:
                if file.endswith(tuple(allowed)):
                    draft_list.append(os.path.join(root, file))
        return draft_list



import os
import re
import configparser
import argparse
from tko.util.decoder import Decoder
from tko.util.text import Text

class RemoteLink:
    def __init__(self):
        self.user = ""
        self.repo = ""
        self.branch = ""
        self.folder = ""
        self.file = ""

    def identify_from_url(self, url: str):
        if url.startswith("https://raw.githubusercontent.com"):
            url = url.replace("https://raw.githubusercontent.com/", "")
            elements = url.split("/")
            self.user = elements[0]
            self.repo = elements[1]
            if elements[2] != "refs":
                raise ValueError("Invalid URL")
            if elements[3] != "heads":
                raise ValueError("Invalid URL")
            self.branch = elements[4]
            self.folder = "/".join(elements[5:-1])
            self.file = elements[-1]
        elif url.endswith(".md"):
            url = url.replace("https://github.com/", "")
            elements = url.split("/")
            self.user = elements[0]
            self.repo = elements[1]
            if elements[2] != "blob":
                raise ValueError("Invalid URL")
            self.branch = elements[3]
            self.folder = "/".join(elements[4:-1])
            self.file = elements[-1]
        else:
            url = url.replace("https://github.com/", "")
            elements = url.split("/")
            self.user = elements[0]
            self.repo = elements[1]
            if elements[2] != "tree":
                raise ValueError("Invalid URL")
            self.branch = elements[3]
            self.folder = "/".join(elements[4:])
        return self

class RemoteCfg:
    def __init__(self, target: str, make_remote: bool):
        self.target = target
        self.remote: RemoteLink = RemoteLink()
        self.cfg_path: str | None = None
        if make_remote:
            self.__load_cfg_path(target)
            self.__parse_cfg()
            if self.cfg_path is None:
                print(Text.format("{r}: remote.cfg file not set", "fail"))

    def cfg_exists(self):
        return self.cfg_path is not None

    def get_cfg_path(self):
        if self.cfg_path is None:
            raise Exception("cfg_path not set")
        return self.cfg_path

    def calc_link_for_local_file(self):
        base_cfg_file = os.path.dirname(os.path.abspath(self.get_cfg_path()))
        base_source_file = os.path.dirname(os.path.abspath(self.target))
        if not base_source_file.startswith(base_cfg_file):
            raise Exception("File not match with remote cfg")
        folder = os.path.relpath(base_source_file, base_cfg_file)
        self.remote.folder = folder
        return self.remote

    def __parse_cfg(self):
        if self.cfg_path is None:
            return
        config = configparser.ConfigParser()
        config.read(self.get_cfg_path())

        self.remote.user = config["DEFAULT"]["user"]
        self.remote.repo = config["DEFAULT"]["rep"]
        self.remote.branch = config["DEFAULT"]["branch"]

    def __load_cfg_path(self, target: str):
        # look for the remote.cfg file in the current folder
        # if not found, look for it in the parent folder
        # if not found, look for it in the parent's parent folder ...

        path = os.path.dirname(os.path.abspath(target))

        while True:
            cfg_path = os.path.join(path, "remote.cfg")
            if os.path.isfile(cfg_path):
                self.cfg_path = cfg_path
                break
            new_path = os.path.dirname(path)
            if new_path == path:
                break
            path = new_path


class Absolute:

    # processa o conteúdo trocando os links locais para links absolutos utilizando a url remota
    @staticmethod
    def __replace_remote(content: str, remote_raw: str, remote_view: str, remote_folder: str) -> str:
        if content == "":
            return ""
        if not remote_raw.endswith("/"):
            remote_raw += "/"
        if not remote_view.endswith("/"):
            remote_view += "/"
        if not remote_folder.endswith("/"):
            remote_folder += "/"

        #trocando todas as imagens com link local
        regex = r"!\[(.*?)\]\((\s*?)([^#:\s]*?)(\s*?)\)"
        subst = r"![\1](" + remote_raw + r"\3)"
        result = re.sub(regex, subst, content, count=0, flags=0)

        regex = r"\[(.+?)\]\((\s*?)([^#:\s]*?)(\s*?/)\)"
        subst = r"[\1](" + remote_folder + r"\3)"
        result = re.sub(regex, subst, result, 0)

        #trocando todos os links locais cujo conteudo nao seja vazio
        regex = r"\[(.+?)\]\((\s*?)([^#:\s]*?)(\s*?)\)"
        subst = r"[\1](" + remote_view + r"\3)"
        result = re.sub(regex, subst, result, 0)

        return result

    @staticmethod
    def relative_to_absolute(content: str, rl: RemoteLink):
        folder = rl.folder
        user_repo = "/".join([rl.user, rl.repo])
        remote_raw    = "/".join(["https://raw.githubusercontent.com", user_repo, rl.branch , folder])
        remote_view    = "/".join(["https://github.com", user_repo, "blob", rl.branch, folder])
        remote_folder = "/".join(["https://github.com", user_repo, "tree", rl.branch, folder])
        return Absolute.__replace_remote(content, remote_raw, remote_view, remote_folder)


    @staticmethod
    def convert_or_copy_or_print(source: str, target: str | None, make_remote: bool = False):
        content = Decoder.load(source)
        cfg = RemoteCfg(source, make_remote)
        if cfg.cfg_exists():
            content = Absolute.relative_to_absolute(content, cfg.calc_link_for_local_file())
        if target is not None:
            Decoder.save(target, content)
        else:
            print(content)
        

def remote_main(args: argparse.Namespace):
    Absolute.convert_or_copy_or_print(args.target, args.output)



from tko.run.diff_builder import DiffBuilder
from tko.run.unit import Unit
from tko.enums.execution_result import ExecutionResult
from tko.util.symbols import symbols
from tko.util.text import Text, Token

class DiffBuilderSide:
    def __init__(self, width: int, unit: Unit):
        self.width = width
        self.curses = False
        self.db = DiffBuilder(width)
        self.unit: Unit = unit
        self.output: list[Text] = []
        self.__to_insert_header = False
        self.__standalone_diff = False
        self.expected_received, self.first_failure = self.db.render_diff(self.unit.get_expected(), self.unit.get_received())

    def set_curses(self):
        self.curses = True
        return self

    def split_screen(self, a: Text | None, b: Text | None, unequal: Token = symbols.vbar) -> Text:
        avaliable = self.width - 7 # 2 spaces before, 2 spaces after, 3 between
        cut = avaliable // 2
        if a is None or b is None or a != b:
            symb = unequal
        else:
            symb = symbols.vbar
        ta = Text() + a
        tb = Text() + b
        ta = ta.ljust(cut, Token(" ")).trim_end(cut)
        tb = tb.ljust(cut, Token(" ")).trim_end(cut)
        line = symb + " " + ta + " " + symb + " " + tb + " "
        if self.width % 2 == 0:
            line += " "
        return line + symbols.vbar

    def title_side_by_side(self, left: Text, right: Text, filler: Token = Token(" "), middle: Token = Token(" "), prefix: Token = Token(), posfix: Token = Token()) -> Text:
        avaliable = self.width - len(prefix) - len(posfix) - len(middle)
        half = avaliable // 2
        line = Text() + prefix

        a = left
        a = a.center(half, filler)
        if len(a) > half:
            a = a.trim_end(half)
        line += a
        line += middle
        b = right
        b = b.center(half, filler)
        if len(b) > half:
            b = b.trim_end(half)
        line += b
        if self.width % 2 == 0:
            line += filler
        line += posfix
        return line

    def to_insert_header(self, value: bool):
        self.__to_insert_header = value
        return self

    def standalone_diff(self):
        self.__standalone_diff = True
        return self

    def _insert_header(self):
        self.output.append(Text().fold_in(self.width, symbols.hbar, "╭", "╮"))
        self.output.append(self.unit.str().fold_in(self.width, " ", "│", "│"))

    def _insert_input(self):
        # input header
        input_color = "b" if self.unit.get_expected() != self.unit.get_received() else "g"
        input_headera = Text().addf(input_color, DiffBuilder.vinput)
        input_headerb = Text().addf(input_color, DiffBuilder.vinput)
        if self.__to_insert_header:
            self.output.append(self.title_side_by_side(input_headera, input_headerb, symbols.hbar, Token("┬"), Token("├"), Token("┤")))
        else:
            self.output.append(Text().addf(input_color, DiffBuilder.vinput).fold_in(self.width, symbols.hbar, "╭", "╮"))

        # input lines
        if self.unit.inserted != "":
            lines = [Text().add(x) for x in self.unit.inserted.splitlines()]
            for l in lines:
                self.output.append(self.split_screen(l, l))
    
    def _insert_expected_received(self):
        # expected and received header
        ecolor = "g" if self.unit.get_expected() == self.unit.get_received() else "y"
        expected_header = Text().addf(ecolor, DiffBuilder.vexpected)
        rcolor = "r" if self.unit.get_expected() != self.unit.get_received() else "g"
        received_header = Text().addf(rcolor, DiffBuilder.vreceived)
        if self.__standalone_diff:
            self.output.append(self.title_side_by_side(expected_header, received_header, symbols.hbar, Token("┬"), Token("╭"), Token("╮")))
        else:
            self.output.append(self.title_side_by_side(expected_header, received_header, symbols.hbar, Token("┼"), Token("├"), Token("┤")))
        # expected and received lines
        symbol = symbols.unequal
        if self.unit.result == ExecutionResult.EXECUTION_ERROR or self.unit.result == ExecutionResult.COMPILATION_ERROR:
            symbol = symbols.vbar
        for exp, rec in self.expected_received:
            self.output.append(self.split_screen(exp, rec, symbol))

    def _insert_first_line_diff(self):
        if self.unit.get_expected() == self.unit.get_received() or self.unit.get_expected() == "":
            return
        self.output.append(Text().addf("b", DiffBuilder.vunequal).fold_in(self.width, symbols.hbar, "├", "┤"))
        for line in self.db.first_failure_diff(self.unit.get_expected(), self.unit.get_received(), self.first_failure):
            width = self.width - 1
            self.output.append(Text().add("│").add(line).ljust(width, Token(" ")).add("│"))
        
    def _finish(self):
        if self.unit.get_expected() == self.unit.get_received() or self.unit.get_expected() == "":
            self.output.append(Text().add("┴").fold_in(self.width, symbols.hbar, "╰", "╯"))
        else:
            self.output.append(Text().fold_in(self.width, symbols.hbar, "╰", "╯"))

    def build_diff(self) -> list[Text]:
        self.output = []
        if self.__to_insert_header:
            self._insert_header()
        if not self.__standalone_diff:
            self._insert_input()
        self._insert_expected_received()
        self._insert_first_line_diff()
        self._finish()
        
        return self.output


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re
import os
import argparse
from tko.util.decoder import Decoder

class TaskLine:
    def __init__(self, line: str = ""):
        self.line: str = line
        self.isTask: bool = False
        self.pre: str = ""
        self.title: str = ""
        self.link: str = ""
        self.pos: str = ""
    
    def initByLine(self):
        regex = r'- \[ \](.*)\[(.+)\]\((.+)\)(.*)'
        match = re.search(regex, self.line)
        if match:
            self.isTask = True
            self.pre = match.group(1)
            self.title = match.group(2)
            self.link = match.group(3)
            self.pos = match.group(4)
        return None

    def initByHook(self, hook: str):
        self.isTask = True
        self.link = os.path.join("base", hook, "Readme.md")
        self.pre = f" `@{hook}` "
        self.load_title_from_link()

    def get_label(self) -> None | str:
        line = self.pre + " " + self.title + " " + self.pos
        if "@" not in line:
            return None
        label = line.split('@')[1]
        valid_chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-"
        valid_label = ""
        for char in label:
            if char in valid_chars:
                valid_label += char
            else:
                break
        if len(valid_label) == 0:
            print("fail: error in", line)
            return None
        return valid_label

    def get_hook(self) -> None | str:
        if self.link.endswith("/Readme.md") and self.link.startswith("base/"):
            try:
                hook = self.link.split("/")[-2]
                return hook
            except Exception as e:
                print("fail: error in", self.link, "with exception:", e)
                exit(1)
        return None

    def load_title_from_link(self) -> str:
        link = self.link
        data = Decoder.load(link)
        header = data.splitlines()[0]
        if (len(header) == 0):
            print('fail: Empty header in ', link)
            exit(1)
        if not header.startswith('# '):
            print('fail: first line should start with # in ', link)
            exit(1)
        words = header.split(' ')
        if len(words) < 2:
            print('fail: header must have any word ', link)
            exit(1)
        return " ".join(words[1:])
    
    def make_new_line(self) -> str:
        if self.isTask:
            label = self.get_label()
            hook = self.get_hook()
            if hook is not None:
                pre = self.pre
                pos = self.pos
                if label is not None:
                    pre = self.pre.replace("@" + label, "@" + hook)
                    pos = self.pos.replace("@" + label, "@" + hook)
                title = self.load_title_from_link()
                return f"- [ ]{pre}[{title}]({self.link}){pos}"
        return self.line

def loading_titles_from_files(path: str) -> list[TaskLine]:
    content = Decoder.load(path)
    lines = content.splitlines()

    output: list[str] = []
    taskLines = [TaskLine(line) for line in lines]
    for taskLine in taskLines:
        taskLine.initByLine()
        line = taskLine.line
        if taskLine.isTask:
            line = taskLine.make_new_line()
        output.append(line)

    Decoder.save(path, '\n'.join(output))
    return taskLines


# def found_labels_mismatch(path: str, base: str) -> bool:
#     error_found = False
#     base = clear_base(base)
#     print("Checking mismatch in labels")
#     data = Decoder.load(path)
#     lines = data.splitlines()

#     not_ok: list[str] = []
#     count_ok = 0
#     for line in lines:
#         hook = get_hook_from_line(line, base)
#         if hook is None:
#             continue
#         label = get_label(line)
#         if label is None:
#             print("fail: error in", line)
#             error_found = True
#             continue
#         if label == hook:
#             count_ok += 1
#         else:
#             not_ok.append("    ({} != {}): {}".format(label, hook, os.path.join(base, hook, 'Readme.md')))
#             error_found = True

#     print("- verified:", count_ok)
#     print("- mismatch:", len(not_ok))
#     for line in not_ok:
#         print(line)
#     return error_found

# check for all folders in the base folder searching for missing labels
def found_unused_hooks(taskLines: list[TaskLine], base_dir: str) -> bool:
    print("Checking for unused hooks")
    hooks_all: list[str] = []
    for tline in taskLines:
        if tline.isTask:
            hook = tline.get_hook()
            if hook is not None:
                hooks_all.append(hook)
    # create a set of hooks
    hooks = set(hooks_all)

    # create a list with folders in base dir if base/folder/Readme.md exists
    folders: list[str] = []
    for folder in os.listdir(base_dir):
        if os.path.isfile(base_dir + '/' + folder + '/Readme.md'):
            folders.append(folder)

    missing = [f for f in folders if f not in hooks]

    if len(missing) > 0:
        print("Missing entries:")
        print(missing)
        for m in missing:
            taskLine = TaskLine()
            taskLine.initByHook(m)
            print(taskLine.make_new_line())
        return True
    return False


def indexer_main(args: argparse.Namespace):
    taskLines: list[TaskLine] = loading_titles_from_files(args.path)
    # if found_labels_mismatch(args.path, args.base):
    #     exit(1)
    if found_unused_hooks(taskLines, args.base):
        exit(1)



from tko.enums.diff_mode import DiffMode
from tko.util.text import Text
from typing import Any
import curses

class AppSettings:

    def __init__(self):
        self.diff_mode = DiffMode.SIDE.value
        self.show_hidden = False
        self.use_images = True
        self.use_borders = False
        self.editor = "code"
        self.timeout = 2
        self.key_left: int = curses.KEY_LEFT
        self.key_right: int = curses.KEY_RIGHT
        self.key_up: int = curses.KEY_UP
        self.key_down: int = curses.KEY_DOWN

    def set_key_left(self, key: int):
        self.key_left = key
        return self
    def set_key_right(self, key: int):
        self.key_right = key
        return self
    def set_key_up(self, key: int):
        self.key_up = key
        return self
    def set_key_down(self, key: int):
        self.key_down = key
        return self
    
    def get_key_left(self) -> int:
        return self.key_left
    def get_key_right(self) -> int:
        return self.key_right
    def get_key_up(self) -> int:
        return self.key_up
    def get_key_down(self) -> int:
        return self.key_down

    def to_dict(self):
        return self.__dict__
    
    def from_dict(self, attr_dict: dict[str, Any]):
        for key, value in attr_dict.items():
            if hasattr(self, key) and type(getattr(self, key)) == type(value):
                setattr(self, key, value)
        return self

    def toggle_diff(self):
        if self.diff_mode == DiffMode.SIDE.value:
            self.diff_mode = DiffMode.DOWN.value
        else:
            self.diff_mode = DiffMode.SIDE.value

    def toggle_borders(self):
        self.use_borders = not self.use_borders
    
    def toggle_images(self):
        self.use_images = not self.use_images
    
    def toggle_hidden(self):
        self.show_hidden = not self.show_hidden

    def set_diff_mode(self, diff_mode: DiffMode):
        self.diff_mode = diff_mode.value
        return self

    def set_show_hidden(self, show_hidden: bool):
        self.show_hidden = show_hidden
        return self

    def set_use_borders(self, borders: bool):
        self.use_borders = borders
        return self
    
    def set_use_images(self, images: bool):
        self.use_images = images
        return self

    def set_editor(self, editor: str):
        self.editor = editor
        return self

    def set_timeout(self, timeout: int):
        self.timeout = timeout
        return self

    def get_diff_mode(self) -> DiffMode:
        if self.diff_mode == DiffMode.SIDE.value:
            return DiffMode.SIDE
        return DiffMode.DOWN

    # def get_lang_default(self) -> str:
    #     return self._lang_default

    # def get_last_rep(self) -> str:
    #     return self._last_rep

    def get_show_hidden(self) -> bool:
        return self.show_hidden

    def get_use_images(self) -> bool:
        return self.use_images

    def get_use_borders(self) -> bool:
        return self.use_borders

    def get_editor(self) -> str:
        return self.editor

    def get_timeout(self) -> int:
        return self.timeout

    def __str__(self):
        output: list[str] = []
        output.append(str(Text.format("{g}", "Configurações globais:")))
        output.append("- Diff    : {}".format(str(self.get_diff_mode().value)))
        output.append("- Editor  : {}".format(self.get_editor()))
        output.append("- Bordas  : {}".format(self.get_use_borders()))
        output.append("- Images  : {}".format(self.get_use_images()))
        output.append("- Timeout : {}".format(self.get_timeout()))
        return "\n".join(output)


# from typing import override
import urllib.request
from tko.feno.remote_md import RemoteLink, Absolute
from tko.util.decoder import Decoder

class RemoteUrl:
    def __init__(self, url: str):
        self.raw_link: str | None = ""
        self.remote: RemoteLink | None = None

        if not url.startswith("https://"):
            raise ValueError("Invalid URL")

        if url.startswith("https://gist.githubusercontent.com"):
            self.raw_link = url
        else:
            try:
                self.remote = RemoteLink().identify_from_url(url)
            except:
                raise Warning("URL inválida para download: {}".format(url))
        self.file = ""

    def get_raw_url(self):
        if self.raw_link:
            return self.raw_link
        if self.remote is None:
            return ""
        return "https://raw.githubusercontent.com/" + self.remote.user + "/" + self.remote.repo + "/" + self.remote.branch + "/" + self.remote.folder + "/" + self.remote.file

    def download_absolute_to(self, filename: str):
        [tempfile, __content] = urllib.request.urlretrieve(self.get_raw_url(), filename)
        content = ""
        content = Decoder.load(tempfile)
        if self.remote is not None:
            content = Absolute.relative_to_absolute(content, self.remote)
        Decoder.save(filename, content)
        return

    # @override
    def __str__(self):
        return self.get_raw_url()



from typing import Any
import os
import urllib
from tko.util.remote_url import RemoteUrl
from tko.game.game import Game
from tko.util.decoder import Decoder
from tko.util.text import Text
import yaml # type: ignore

available_languages = ["c", "cpp", "py", "ts", "java", "go", "hs", "yaml"]

def remove_git_merge_tags(lines: list[str]) -> list[str]:
    # remove lines with <<<<<<<, =======, >>>>>>>>
    filtered_lines: list[str] = []
    for line in lines:
        if line.startswith("<<<<<<<") or line.startswith("=======") or line.startswith(">>>>>>>"):
            continue
        filtered_lines.append(line)
    return filtered_lines

class Repository:
    CFG_FILE = "repository.yaml"
    GLOBAL_LOG_FILE = "history.csv"
    TASK_LOG_FILE = "task_log.csv"
    DAILY_FILE = "daily.yaml"
    INDEX_FILE = "Readme.md"
    TRACK_FOLDER = "track"
    CONFIG_FOLDER = ".tko"

    def __init__(self, folder: str):
        self.root_folder: str = folder
        rec_folder = Repository.rec_search_for_repo(folder)
        if rec_folder != "":
            self.root_folder = rec_folder

        self.data: dict[str, Any] = {}
        self.game = Game()
        self.upgrade_version()

    def upgrade_version(self):
        return self

    @staticmethod
    def rec_search_for_repo(folder: str) -> str:
        folder = os.path.abspath(folder)
        if os.path.exists(os.path.join(folder, Repository.CONFIG_FOLDER, Repository.CFG_FILE)):
            return folder
        new_folder = os.path.dirname(folder)
        if new_folder == folder: # não encontrou, não tem mais como subir
            return ""
        return Repository.rec_search_for_repo(new_folder)

    def get_key_from_task_folder(self, folder: str) -> str:
        label = os.path.basename(folder)
        task_folder = self.get_task_folder_for_label(label)
        if task_folder == folder:
            return label
        return ""

    def is_task_folder(self, folder: str) -> bool:
        label = os.path.basename(folder)
        task_folder = self.get_task_folder_for_label(label)
        return task_folder == folder

    def get_track_folder(self) -> str:
        return os.path.abspath(os.path.join(self.root_folder, Repository.CONFIG_FOLDER, Repository.TRACK_FOLDER))

    def get_track_task_folder(self, label: str) -> str:
        return os.path.abspath(os.path.join(self.root_folder, Repository.CONFIG_FOLDER, Repository.TRACK_FOLDER, label))

    def get_task_folder_for_label(self, label: str) -> str:
        return os.path.abspath(os.path.join(self.root_folder, self.get_database_folder(), label))

    def is_local_dir(self, path: str) -> bool:
        rep_dir = self.get_rep_dir()
        path = os.path.abspath(path)
        return os.path.commonpath([rep_dir, path]) == rep_dir

    @staticmethod
    def is_web_link(link: str) -> bool:
        return link.startswith("http:") or link.startswith("https:")

    def load_game(self):
        if self.data == {}:
            self.load_config()
        database_folder = os.path.join(self.root_folder, self.get_database_folder())
        cache_or_index = self.load_index_or_cache()
        self.game.parse_file_and_folder(cache_or_index, database_folder, self.get_lang())
        self.__load_tasks_from_rep_into_game()
        return self
    
    def __load_tasks_from_rep_into_game(self):
        # load tasks from repository.yaml into game
        tasks = self.get_tasks()
        for key, serial in tasks.items():
            if key in self.game.tasks:
                self.game.tasks[key].load_from_db(serial)

    def get_config_file(self) -> str:
        return os.path.abspath(os.path.join(self.root_folder, Repository.CONFIG_FOLDER, Repository.CFG_FILE))
    
    def get_config_backup_file(self) -> str:
        return self.get_config_file() + ".backup"
    
    def get_history_file(self) -> str:
        return os.path.abspath(os.path.join(self.root_folder, Repository.CONFIG_FOLDER, Repository.GLOBAL_LOG_FILE))
    
    def get_daily_file(self) -> str:
        return os.path.abspath(os.path.join(self.root_folder, Repository.CONFIG_FOLDER, Repository.DAILY_FILE))
    
    def get_default_readme_path(self) -> str:
        return os.path.abspath(os.path.join(self.root_folder, Repository.INDEX_FILE))

    def get_rep_dir(self) -> str:
        return os.path.abspath(self.root_folder)

    def __load_local_file(self, source: str) -> str:
        # test if file is inside or outside the repository
        source = os.path.abspath(os.path.join(self.root_folder, self.CONFIG_FOLDER, source))
        if not os.path.exists(source):
            raise Warning(Text.format("{r}: Arquivo fonte do repositório {y} não foi encontrado", "Erro", source))
        return source

    def __is_remote_source(self) -> bool:
        source = self.get_rep_source()
        return source.startswith("http:") or source.startswith("https:")
    
    def down_source_from_remote_url(self):
        source = self.get_rep_source()
        cache_file = self.get_default_readme_path()
        os.makedirs(self.root_folder, exist_ok=True)
        ru = RemoteUrl(source)
        try:
            ru.download_absolute_to(cache_file)
        except urllib.error.URLError: # type: ignore
            print("Não foi possível baixar o arquivo do repositório")
            if os.path.exists(cache_file):
                print("Usando arquivo do cache")
            else:
                raise Warning("fail: Arquivo do cache não encontrado")
        return cache_file

    def load_index_or_cache(self) -> str:
        source = self.get_rep_source()

        if self.__is_remote_source():
            return self.down_source_from_remote_url()
        return self.__load_local_file(source)

    __version = "version"
    __actual_version = "0.0.1"
    __source = "remote"  #remote url or local file or remote file
    __expanded = "expanded"
    __new_items = "new_items"
    __tasks = "tasks"
    __flags = "flags"
    __lang = "lang"
    __selected = "index"
    __database = "database"

    defaults: dict[str, Any] = {
        __version: __actual_version,
        __database: "database",
        __source: "",
        __expanded: [],
        __tasks: {},
        __flags: {},
        __new_items: [],
        __lang: "",
        __selected: ""
    }

    def get_database_folder(self) -> str:
        return self.__get(Repository.__database)

    def set_database_folder(self, value: str):
        self.__set(Repository.__database, value)
        return self

    def get_version(self) -> str:
        return self.__get(Repository.__version)
    
    def set_version(self, value: str):
        self.__set(Repository.__version, value)
        return self

    def get_rep_source(self) -> str:
        return self.__get(Repository.__source)
    
    def set_rep_source(self, value: str):
        self.__set(Repository.__source, value)
        return self

    def get_selected(self) -> str:
        return self.__get(Repository.__selected)

    def get_expanded(self) -> list[str]:
        return self.__get(Repository.__expanded)

    def get_new_items(self) -> list[str]:
        return self.__get(Repository.__new_items)
    
    def get_tasks(self) -> dict[str, Any]:
        return self.__get(Repository.__tasks)
    
    def get_flags(self) -> dict[str, Any]:
        return self.__get(Repository.__flags)
    
    def get_lang(self) -> str:
        return self.__get(Repository.__lang)

    def set_expanded(self, value: list[str]):
        return self.__set(Repository.__expanded, value)
    
    def set_new_items(self, value: list[str]):
        return self.__set(Repository.__new_items, value)
    
    def set_tasks(self, value: dict[str, str]):
        self.__set(Repository.__tasks, value)
        return self
    
    def set_flags(self, value: dict[str, Any]):
        self.__set(Repository.__flags, value)
        return self
    
    def set_lang(self, value: str):
        self.__set(Repository.__lang, value)
        return self
    
    def set_selected(self, value: str):
        self.__set(Repository.__selected, value)
        return self

    def __get(self, key: str) -> Any:
        if key not in self.defaults:
            raise ValueError(f"Key {key} not found in RepSettings")
        value = self.data.get(key, Repository.defaults[key])
        if type(value) != type(Repository.defaults[key]):
            return Repository.defaults[key]
        return value

    def __set(self, key: str, value: Any):
        self.data[key] = value
        return self

    def create_empty_config_file(self, point_to_local_readme: bool = False):
        if point_to_local_readme:
            self.set_rep_source(self.get_default_readme_path())
        self.save_config()
        return self

    def has_local_config_file(self) -> bool:
        return os.path.exists(self.get_config_file())

    # config
    def load_config(self):
        encoding = Decoder.get_encoding(self.get_config_file())
        try:
            with open(self.get_config_file(), encoding=encoding) as f:
                lines = f.readlines()
                lines = remove_git_merge_tags(lines)
                self.data = yaml.safe_load("\n".join(lines))
            if self.data is None or not isinstance(self.data, dict) or len(self.data) == 0: # type: ignore
                with open(self.get_config_backup_file(), "r", encoding=encoding) as f:
                    lines = f.readlines()
                    lines = remove_git_merge_tags(lines)
                    self.data = yaml.safe_load("\n".join(lines))
            if self.data is None or not isinstance(self.data, dict) or len(self.data) == 0: # type: ignore
                raise FileNotFoundError(f"Arquivo de configuração vazio: {self.get_config_file()}")
        except:
            raise Warning(Text.format("O arquivo de configuração do repositório {y} está {r}.\nAbra e corrija o conteúdo ou crie um novo.", self.get_config_file(), "corrompido"))
        return self

    def save_config(self):
        self.set_version(self.__actual_version)
        yaml_file = self.get_config_file()
        if not os.path.exists(os.path.dirname(yaml_file)):
            os.makedirs(os.path.dirname(yaml_file))
        # filter keys that are not in defaults
        for key in list(self.data.keys()):
            if key not in Repository.defaults:
                del self.data[key]

        for key in Repository.defaults:
            if key not in self.data:
                self.data[key] = Repository.defaults[key]
        yaml_file_new = yaml_file + ".new"
        yaml_file_bkp = yaml_file + ".backup"
        with open(yaml_file_new, "w", encoding="utf-8") as f:
            yaml.dump(self.data, f, default_flow_style=False, allow_unicode=True, sort_keys=False)
        # copy file from yaml_file to yaml_file.backup
        try:
            Repository.rename_file(yaml_file, yaml_file_bkp)
        except:
            pass
        Repository.rename_file(yaml_file_new, yaml_file)
        return self

    def __str__(self) -> str:
        return ( f"data: {self.data}\n" )

    @staticmethod
    def rename_file(old_name: str, new_name: str):
        if os.path.exists(old_name):
            if os.path.exists(new_name):
                os.remove(new_name)
            os.rename(old_name, new_name)
        else:
            raise Warning(Text.format("O arquivo {y} não foi encontrado", old_name))


import yaml # type: ignore
from tko.settings.task_basic import TaskBasic
from tko.settings.log_action import LogAction
from tko.settings.log_info import LogInfo

class DailyListener:
    def __init__(self):
        self.history: dict[str, dict[str, TaskBasic]] = {} # dict[day, dict[key, [key, cov, how]]]
        self.actual: dict[str, TaskBasic] = {}
        self.daily_file: str | None
        
    def listener(self, action: LogAction, new_entry: bool = False):
        decoder = LogInfo().decode(action)
        types = [LogAction.Type.TEST.value, LogAction.Type.PROG.value, LogAction.Type.SELF.value]
        if decoder.type in types:
            self.log_task(decoder.timestamp, decoder.key, decoder.coverage, decoder.approach, decoder.autonomy)

    def set_daily_file(self, daily_file: str):
        self.daily_file = daily_file
        return self

    def log_task(self, timestamp: str, key: str, coverage: int = -1, autonomy: int = -1, skill: int = -1):
        if key not in self.actual:
            self.actual[key] = TaskBasic(key)
        self.actual[key].set_coverage(coverage).set_approach(autonomy).set_autonomy(skill)

        day = timestamp.split(" ")[0]
        if day not in self.history:
            self.history[day] = {}
        if key not in self.history[day]:
            self.history[day][key] = self.actual[key]
        else:
            self.history[day][key].set_coverage(coverage).set_approach(autonomy).set_autonomy(skill)
        # self.save_yaml()

    def __str__(self) -> str:
        output: dict[str, dict[str, str]] = {}
        for day in sorted(self.history.keys()):
            output[day] = {}
            for key in self.history[day]:
                output[day][key] = str(self.history[day][key])
        return yaml.dump(output)
    
    def save_yaml(self):
        if self.daily_file is None:
            return
        with open(self.daily_file, 'w') as file:
            file.write(str(self))



from tko.settings.repository import Repository
from tko.util.text import Text, Token
from tko.util.to_asc import SearchAsc
from tko.play.flags import Flags
from tko.game.cluster import Cluster
from tko.game.quest import Quest
from tko.game.task import Task
from tko.play.border import Border
from tko.down.drafts import Drafts
from tko.util.symbols import symbols
from tko.settings.settings import Settings
from tko.play.floating_manager import FloatingManager
from tko.util.symbols import symbols
from tko.game.tree_item import TreeItem
from tko.settings.logger import Logger


class TaskAction:
    BAIXAR   = "Baixar  "
    EXECUTAR = "Escolher"
    VISITAR  = "Visitar "
    EXPANDIR = "Expandir"
    CONTRAIR = "Contrair"
    BLOQUEIO = "Travado"

class TaskTree:

    def __init__(self, settings: Settings, rep: Repository, fman: FloatingManager):
        self.settings = settings
        self.app = settings.app
        self.game = rep.game
        self.rep = rep
        self.fman = fman
        self.style = Border(settings.app)
        self.colors = settings.colors
        self.items: list[TreeItem] = []
        self.all_items: dict[str, TreeItem] = {}
        self.selected_item: str = ""
        self.index_begin: int = 0
        self.max_title: int = 0
        self.search_text: str = ""
        self.load_all_items()
        self.load_from_rep()
        self.update_tree(admin_mode = Flags.admin.is_true(), first_loop=True)
        self.reload_sentences()

    def load_all_items(self):
        for c in self.game.clusters.values():
            self.all_items[c.key] = c
            for q in c.get_quests():
                self.all_items[q.key] = q
                for t in q.get_tasks():
                    self.all_items[t.key] = t

    def load_from_rep(self):
        # self.new_items: list[str] = [v for v in self.rep.get_new_items()]
        self.expanded: list[str] = [v for v in self.rep.get_expanded()]
        self.selected_item = self.rep.get_selected()

        # tasks = self.rep.get_tasks()
        # for key, serial in tasks.items():
        #     if key in self.game.tasks:
        #         self.game.tasks[key].load_from_db(serial)

    def save_on_rep(self):
        self.rep.set_expanded(self.expanded)
        # self.rep.set_new_items([x for x in set(self.new_items)])
        self.rep.set_selected(self.selected_item)
        tasks: dict[str, str] = {}
        for t in self.game.tasks.values():
            if not t.is_db_empty():
                tasks[t.key] = t.save_to_db()
        self.rep.set_tasks(tasks)

    def update_tree(self, admin_mode: bool, first_loop: bool = False):
        if not admin_mode:
            # reachable_clusters = [c.key for c in self.game.clusters.values() if c.is_reachable()]
            reachable_quests = [q.key for q in self.game.quests.values() if q.is_reachable()]
            # old_reachable = reachable_clusters + reachable_quests

        self.game.update_reachable_and_available()
            
        if not admin_mode:
            # reachable_clusters = [c.key for c in self.game.clusters.values() if c.is_reachable()]
            reachable_quests = [q.key for q in self.game.quests.values() if q.is_reachable()]
            reachable_keys = reachable_quests + list(self.game.clusters.keys())
            for key in self.expanded:
                if key not in reachable_keys:
                    self.expanded.remove(key)
            # if not first_loop:
            #     for key in reachable_keys:
            #         if key not in old_reachable and key not in self.new_items:
            #             self.new_items.append(key)

        # remove expanded items from new
        # self.new_items = [item for item in self.new_items if item not in self.expanded]

    def __update_max_title(self):
        min_value = 50
        items: list[int] = []
        for c in self.game.clusters.values():
            if c.key in self.game.clusters.keys():
                items.append(len(c.title))
                if c.key in self.expanded:
                    for q in c.get_quests():
                        items.append(len(q.get_full_title()) + 2)
                        if q.key in self.expanded:
                            for t in q.get_tasks():
                                items.append(len(t.title) + 6)

        self.max_title = min_value
        if len(items) > 0:
            self.max_title = max(items)
        if self.max_title < min_value:
            self.max_title = min_value

    def __str_task(self, focus_color: str, t: Task, lig_cluster: str, lig_quest: str, quest_reachable: bool) -> Text:
        down_symbol = Token(" ")
        in_focus = focus_color != ""
        down_symbol = symbols.task_to_visit
        if t.link_type == Task.Types.STATIC_FILE:
            down_symbol = symbols.task_local
        if t.link_type == Task.Types.REMOTE_FILE or t.link_type == Task.Types.IMPORT_FILE:
            down_symbol = symbols.task_to_download
            if self.is_downloaded_for_lang(t):
                down_symbol = symbols.task_downloaded
    
        color_aval = "g" if quest_reachable and t.is_reachable() else "r"
        color_lig_task = color_aval
        if self.app.get_show_hidden() == False and not Flags.admin:
            quest = self.game.quests[t.quest_key]
            if quest.prog:
                if t.key != self.game.quests[t.quest_key].get_tasks()[-1].key:
                    if t.get_percent() < 50:
                        color_lig_task = "y"
                    

        output = Text()
        output.add(" ").addf(color_aval, lig_cluster)
        output.add(" ")
        output.addf(color_lig_task, lig_quest)
        output.add(down_symbol).add(" ")
        output.add(t.get_prog_symbol()).add(" ")
        output.add(symbols.approach_list[t.approach]).add(" ")
        output.add(symbols.autonomy_list[t.autonomy])


        if in_focus:
            output.add(self.style.roundL(focus_color))
        else:
            output.add(" ")

        color = ""
        if in_focus:
            color = "k" + focus_color

        for word in t.title.split(" "):
            if word.startswith("@") or word.startswith("#") or word.startswith("!"):
                output.addf(color + "g", word + " ")
            elif word.startswith(":"):
                output.addf(color + "y", word + " ")
            elif word.startswith("*"):
                output.addf(color + "c", word + " ")
            elif word.startswith("+"):
                output.addf(color + "c", word + " ")
            else:
                output.addf(color, word + " ")

        if in_focus:
            output.add(self.style.roundR(focus_color))
        else:
            output.add(" ")

        output.ljust(self.max_title + 10, Token(" "))
        if Flags.percent.is_true():
            prog = int(t.get_percent())
            output.addf("y", str(prog).rjust(3, " ") + "%")
        else:
            color_rate: list[str] = ["", "c", "g", "y", "m", "r"]
            rate = (t.howclear + t.howfun) / 2
            for i in range(1, 6):
                if rate >= i:
                    output.addf(color_rate[t.howeasy], symbols.star)
                else:
                    output.addf(color_rate[t.howeasy], symbols.open_star)

        # if Flags.reward:
        #     xp = ""
        #     for s, v in t.skills.items():
        #         xp += f" +{s}:{v}"
        #     output.addf(self.colors.task_skills, xp)
            
        return output

    def __str_quest(self, has_kids: bool, focus_color: str, q: Quest, lig: str) -> Text:
        con = "━─"
        if q.key in self.expanded and has_kids:
            con = "─┯"

        color_reachable = "g" if q.is_reachable() else "r"
        if self.app.get_show_hidden() == False and not Flags.admin:
            for quest in self.game.quests.values():
                if quest.is_reachable() == False:
                    if q.key in quest.requires:
                        color_reachable = "y"
                        break

        output: Text = Text().addf(color_reachable, " " + lig + con)

        in_focus = focus_color != ""

        if focus_color == "":
            item_key = self.selected_item
            if item_key in list(self.game.quests.keys()):
                quest = self.game.quests[item_key]
                if not quest.is_reachable():
                    if q.key in quest.requires:
                        focus_color = "y"
                # elif quest.key in q.requires:
                #     focus_color = "r"

        if in_focus:
            output.add(self.style.roundL(focus_color))
        else:
            output.add(" ")

        # color = ""
        # if in_focus:
        #     color = "k" + focus_color

        title = q.get_full_title()
        title = title.ljust(self.max_title + 3, ".")

        output.addf(focus_color, title)

        # done = color + self.colors.task_text_done
        # todo = color + self.colors.task_text_todo
        # output.add(self.style.build_bar(title, q.get_percent() / 100, len(title), done, todo, round=False))

        if in_focus:
            output.add(self.style.roundR(focus_color))
        else:
            output.add(" ")

        # if Flags.percent:
        output.add(" ").add(q.get_resume_by_percent())
        # else:
        #     output.add(" ").add(q.get_resume_by_tasks())

        # if Flags.minimum:
        #     output.add(" ").add(q.get_requirement())

        # if Flags.reward:
        #     xp = ""
        #     for s, v in q.skills.items():
        #         xp += f" +{s}:{v}"
        #     output.addf(self.colors.task_skills, xp)

        # if q.key in self.new_items:
        #     output.addf(self.colors.task_new, " [new]")

        return output

    def __str_cluster(self, has_kids: bool, focus_color: str, cluster: Cluster) -> Text:
        output: Text = Text()
        opening = "━─"
        if cluster.key in self.expanded and has_kids:
            opening = "─┯"
        color_reachable = "g" if cluster.is_reachable() else "r"
        output.addf(color_reachable, opening)

        color = ""
        if focus_color != "":
            color = "k" + focus_color
        title = cluster.title

        title = cluster.title.ljust(self.max_title + 5, ".")
        if focus_color != "":
            output.add(self.style.roundL(focus_color))
        else:
            output.add(" ")

        output.addf(color, title)
        # done = color + self.colors.task_text_done
        # todo = color + self.colors.task_text_todo
        # output.add(self.style.build_bar(title, cluster.get_percent() / 100, len(title), done, todo, round=False))

        if focus_color != "":
            output.add(self.style.roundR(focus_color))
        else:
            output.add(" ")


        # if Flags.percent:
        output.add(" ").add(cluster.get_resume_by_percent())
        # else:
        #     output.add(" ").add(cluster.get_resume_by_quests())
        # if cluster.key in self.new_items:
        #     output.addf(self.colors.task_new, " [new]")

        return output

    def __get_focus_color_cluster(self, item: Cluster) -> str:
        if not item.is_reachable():
            return "R"
        return self.colors.focused_item
    
    def __get_focus_color_quest(self, item: Quest) -> str:
        if not item.is_reachable():
                return "R"
        return self.colors.focused_item

    def filter_by_search(self) -> tuple[set[str], str | None]:
        matches: set[str] = set()
        search = SearchAsc(self.search_text)
        first: None | str = None
        for cluster in self.game.clusters.values():
            if search.inside(cluster.title):
                matches.add(cluster.key)
                first = first or cluster.key
            for quest in cluster.get_quests():
                if search.inside(quest.title):
                    first = first or quest.key
                    matches.add(cluster.key)
                    matches.add(quest.key)
                for task in quest.get_tasks():
                    if search.inside(task.title):
                        first = first or task.key
                        matches.add(cluster.key)
                        matches.add(quest.key)
                        matches.add(task.key)
        return (matches, first)

    def __try_add(self, filtered: set[str], matcher: SearchAsc, item: TreeItem):
        if self.search_text == "":
            self.items.append(item)
            return True
        if item.key in filtered:
            pos = matcher.find(item.sentence.get_str())
            found = pos != -1
            if found:
                for i in range(pos, pos + len(self.search_text)):
                    item.sentence.data[i].fmt = "Y"
            self.items.append(item)
            return True
        return False

    def reload_sentences(self):
        self.__update_max_title()
        self.items = []

        filtered, _ = self.filter_by_search()
        matcher = SearchAsc(self.search_text)

        hide = not self.app.get_show_hidden() and not Flags.admin

        clusters = [self.game.clusters[key] for key in self.game.clusters.keys() if key in filtered]
        if hide:
            clusters = [c for c in clusters if c.is_reachable()]

        for cluster in clusters:
            quests = [q for q in cluster.get_quests() if q.key in self.game.quests.keys() if q.key in filtered]
            if hide:
                quests = [q for q in quests if q.is_reachable()]
            focus_color = self.__get_focus_color_cluster(cluster) if self.selected_item == cluster.get_key() else ""
            cluster.sentence = self.__str_cluster(len(quests) > 0, focus_color, cluster)

            self.__try_add(filtered, matcher, cluster)

            if cluster.key not in self.expanded:  # adicionou o cluster, mas não adicione as quests
                continue

            for q in quests:
                tasks =[t for t in q.get_tasks() if t.key in filtered]
                if hide:
                    tasks = [t for t in tasks if t.is_reachable()]
                lig = "├" if q != quests[-1] else "╰"
                quest_focus_color: str = self.__get_focus_color_quest(q)
                focus_color = quest_focus_color if self.selected_item == q.get_key() else ""
                q.sentence = self.__str_quest(len(tasks) > 0, focus_color, q, lig)

                # self.items.append(Entry(q, sentence))
                self.__try_add(filtered, matcher, q)
                if q.key in self.expanded:
                    for t in tasks:
                        ligc = "│" if q != quests[-1] else " "
                        ligq = "├ " if t != tasks[-1] else "╰ "
                        focus_color: str = quest_focus_color if self.selected_item == t.get_key() else ""
                        t.sentence: Text = self.__str_task(focus_color, t, ligc, ligq, q.is_reachable()) # type: ignore
                        self.__try_add(filtered, matcher, t)
        # verifying if has any selected item
        if self.items:
            found = False
            for item in self.items:
                if item.get_key() == self.selected_item:
                    found = True
                    break
            if found == False:
                self.selected_item = self.items[0].get_key()
                self.reload_sentences()

    def process_collapse_all(self):
        if any([q in self.expanded for q in self.game.quests.keys()]):
            self.expanded = [key for key in self.expanded if key not in self.game.quests.keys()]
        else:
            self.expanded = []

    def process_expand_all(self):
        # if any cluster outside expanded
        expand_clusters = False
        for ckey in self.game.clusters.keys():
            if ckey not in self.expanded:
                expand_clusters = True
        if expand_clusters:
            for ckey in self.game.clusters.keys():
                if ckey not in self.expanded:
                    self.expanded.append(ckey)
        else:
            for qkey in self.game.quests.keys():
                if qkey not in self.expanded:
                    self.expanded.append(qkey)

    def  get_selected_index(self) -> int:
        for i, item in enumerate(self.items):
            if item.key == self.selected_item:
                return i
        return 0

    def get_selected_throw(self) -> TreeItem:
        if len(self.items) > 0:
            index = self.get_selected_index()
            return self.items[index]
        raise IndexError("No item selected")
            
    def mass_mark(self):
        obj = self.get_selected_throw()
        if isinstance(obj, Cluster):
            cluster: Cluster = obj
            if cluster.key not in self.expanded:
                self.expanded.append(cluster.key)
                return
            full_open = True
            for q in cluster.get_quests():
                if q.key in self.game.quests.keys() and q.key not in self.expanded:
                    self.expanded.append(q.key)
                    full_open = False
            if not full_open:
                return

            value = None
            for q in obj.get_quests():
                for t in q.get_tasks():
                    if value is not None:
                        self.set_grade(t, value[0], value[1], value[2])
                    else:
                        value = (100, 5, 5) if t.get_percent() < 100 else (0, 0, 0)
                        self.set_grade(t, value[0], value[1], value[2])
        elif isinstance(obj, Quest):
            if obj.key not in self.expanded:
                self.expanded.append(obj.key)
            else:
                value = None
                for t in obj.get_tasks():
                    if value is not None:
                        self.set_grade(t, value[0], value[1], value[2])
                    else:
                        value = (100, 5, 5) if t.get_percent() < 100 else (0, 0, 0)
                        self.set_grade(t, value[0], value[1], value[2])
        elif isinstance(obj, Task):
            value = (100, 5, 5) if obj.get_percent() < 100 else (0, 0, 0)
            self.set_grade(obj, value[0], value[1], value[2])

    def set_grade(self, task: Task, coverage: int, autonomy: int, skill: int):
        obj = task
        Logger.get_instance().record_self_grade(obj.key, coverage, autonomy, skill)
        obj.set_coverage(coverage)
        obj.set_approach(autonomy)
        obj.set_autonomy(skill)


    def set_selected_by_index(self, index: int):
        if index < 0:
            index = 0
        if index >= len(self.items):
            index = len(self.items) - 1
        self.selected_item = self.items[index].key

    def arrow_right(self):
        if len(self.items) == 0:
            return
        index = self.get_selected_index()
        obj = self.items[index]
        if isinstance(obj, Cluster):
            if not self.unfold(obj):
                self.set_selected_by_index(index + 1)
        elif isinstance(obj, Quest):
            if not self.unfold(obj):
                while True:
                    index += 1
                    if index >= len(self.items):
                        break
                    obj = self.items[index]
                    if isinstance(obj, Cluster) or isinstance(obj, Quest):
                        break
                self.set_selected_by_index(index)
        elif isinstance(obj, Task):
            while True:
                obj = self.items[index]
                if isinstance(obj, Quest) or isinstance(obj, Cluster):
                    break
                if index == len(self.items) - 1:
                    break
                index += 1
            self.set_selected_by_index(index)

    def walk_left_on_quest(self, index: int):
        while True:
            if index == 0:
                break
            index -= 1
            obj = self.items[index]
            if isinstance(obj, Cluster) or isinstance(obj, Quest) and obj.key in self.expanded:
                break
            if index == 0:
                break
        self.set_selected_by_index(index)

    def walk_left_on_cluster(self, index: int):
        while True:
            if index == 0:
                break
            index -= 1
            obj = self.items[index]
            if isinstance(obj, Cluster) or isinstance(obj, Quest):
                break
            if index == 0:
                break
        self.set_selected_by_index(index)

    def arrow_left(self):
        if len(self.items) == 0:
            return
        index = self.get_selected_index()
        obj = self.items[index]
        if isinstance(obj, Quest):
            if not self.fold(obj):
                self.walk_left_on_quest(index)
            return
        elif isinstance(obj, Cluster):
            if obj.key in self.expanded:
                self.expanded.remove(obj.key)
                for q in obj.get_quests():
                    try:
                        self.expanded.remove(q.key)
                    except ValueError:
                        pass
            else:
                self.walk_left_on_cluster(index)
            return
        elif isinstance(obj, Task):
            while True:
                obj = self.items[index]
                if isinstance(obj, Quest):
                    break
                index -= 1
            self.set_selected_by_index(index)

    def unfold(self, obj: TreeItem) -> bool:
        if isinstance(obj, Quest) and not obj.is_reachable() and not Flags.admin:
            return False
        if isinstance(obj, Quest) or isinstance(obj, Cluster):
            if obj.key not in self.expanded:
                self.expanded.append(obj.key)
                if isinstance(obj, Cluster):
                    cluster: Cluster = obj
                    if len(cluster.get_quests()) == 1:
                        self.expanded.append(cluster.get_quests()[0].key)
                return True
        return False

    def fold(self, obj: TreeItem) -> bool:
        if isinstance(obj, Quest) or isinstance(obj, Cluster):
            if obj.key in self.expanded:
                self.expanded.remove(obj.key)
                if isinstance(obj, Quest):
                    quest: Quest = obj
                    cluster = self.game.clusters[quest.cluster_key]
                    if len(cluster.get_quests()) == 1:
                        self.expanded.remove(cluster.key)
                        self.selected_item = cluster.key
                return True
        return False

    def toggle(self, obj: Quest | Cluster):
            if not self.fold(obj):
                self.unfold(obj)

    def get_senteces(self, dy: int) -> list[Text]:
        index = self.get_selected_index()
        if len(self.items) < dy:
            self.index_begin = 0
        else:
            if index < self.index_begin:  # subiu na tela
                self.index_begin = index
            elif index >= dy + self.index_begin:  # desceu na tela
                self.index_begin = index - dy + 1

        sentences: list[Text] = []
        for i in range(self.index_begin, len(self.items)):
            sentences.append(self.items[i].sentence)
        return sentences

    def move_up(self):
        if len(self.items) == 0:
            return
        index = self.get_selected_index()
        self.set_selected_by_index(index - 1)

    def move_down(self):
        if len(self.items) == 0:
            return
        index = self.get_selected_index()
        self.set_selected_by_index(min(len(self.items) - 1, index + 1))

    # return a TaskAction
    def get_task_action(self, task: Task) -> tuple[str, str]:
        if task.link_type == Task.Types.VISITABLE_URL:
            return ("B", TaskAction.VISITAR)
        if task.link_type == Task.Types.STATIC_FILE:
            return ("G", TaskAction.EXECUTAR)
        
        if not self.is_downloaded_for_lang(task):
            return ("Y", TaskAction.BAIXAR)
        return ("G", TaskAction.EXECUTAR)
        
    def is_downloaded_for_lang(self, task: Task):
        folder = task.get_folder()
        lang = self.rep.get_lang()
        if folder is None:
            return False
        drafts = Drafts.load_drafts_only(folder, lang)
        if drafts:
            return True
        return False
        


from tko.game.cluster import Cluster
from tko.game.quest import Quest
from tko.game.task import Task

from tko.util.decoder import Decoder
from tko.game.game_builder import GameBuilder
from tko.game.game_validator import GameValidator
# from typing import override

import yaml # type: ignore
import re

def load_html_tags(task: str) -> None | str:
    pattern = r"<!--\s*(.*?)\s*-->"
    match = re.search(pattern, task)
    if not match:
        return None
    return match.group(1).strip()


class Game:
    def __init__(self):
        self.ordered_clusters: list[str] = [] # ordered clusters
        self.clusters: dict[str, Cluster] = {} 
        self.quests: dict[str, Quest] = {}  # quests indexed by quest key
        self.tasks: dict[str, Task] = {}  # tasks indexed by task key

        self.token_level_one = "level_one"
        self.token_level_mult = "level_mult"
        self.level_one: int = 100
        self.level_mult: float = 1.5


    def parse_xp(self, content: str):
        if content.startswith('---'):
            front_matter = content.split('---')[1].strip()
            yaml_data = yaml.safe_load(front_matter)
            if self.token_level_one in yaml_data:
                self.level_one = int(yaml_data[self.token_level_one])
            if self.token_level_mult in yaml_data:
                self.level_mult = float(yaml_data[self.token_level_mult])

    def get_task(self, key: str) -> Task:
        if key in self.tasks:
            return self.tasks[key]
        raise Warning(f"fail: tarefa '{key}' não encontrada no curso")

    def get_xp_resume(self):
        total = 0
        obtained = 0
        for q in self.quests.values():
            o, t = q.get_xp()
            total += t
            obtained += o
        return obtained, total

    def get_skills_resume(self, avaliable_quests: list[Quest]) -> tuple[dict[str, int], dict[str, int]]:
        total: dict[str, int] = {}
        obtained: dict[str, int] = {}
        avaliable_keys = [q.key for q in avaliable_quests]
        for q in self.quests.values():
            reachable = q.key in avaliable_keys
            for t in q.get_tasks():
                for s in t.skills:
                    if s in total:
                        total[s] += t.skills[s]
                        if reachable:
                            obtained[s] += int(t.skills[s] * t.get_ratio())
                        else:
                            obtained[s] += 0
                    else:
                        total[s] = t.skills[s]
                        if reachable:
                            obtained[s] = int(t.skills[s] * t.get_ratio())
                        else:
                            obtained[s] = 0
                for s in t.qskills:
                    if s in total:
                        total[s] += t.qskills[s]
                        if reachable:
                            obtained[s] += int(t.qskills[s] * t.get_ratio())
                        else:
                            obtained[s] += 0
                    else:
                        total[s] = t.qskills[s]
                        if reachable:
                            obtained[s] = int(t.qskills[s] * t.get_ratio())
                        else:
                            obtained[s] = 0
        return total, obtained


    def parse_file_and_folder(self, filename: str, folder: str, language: str):
        self.filename = filename
        content = Decoder.load(filename)
        self.parse_xp(content)

        gb = GameBuilder(filename, folder).build_from(content, language)
        self.ordered_clusters = gb.ordered_clusters
        self.clusters = gb.clusters
        self.quests = gb.collect_quests()
        self.tasks = gb.collect_tasks()
        GameValidator(filename, self.clusters, self.quests, self.tasks).validate()

        # for t in self.tasks.values():
        #     t.get_link(os.path.dirname(filename) + "/")

    @staticmethod
    def is_reachable_quest(q: Quest, cache: dict[str, bool]):
        if q.key in cache:
            return cache[q.key]

        if len(q.requires_ptr) == 0:
            cache[q.key] = True
            return True
        cache[q.key] = all([r.is_complete() and Game.is_reachable_quest(r, cache) for r in q.requires_ptr])
        return cache[q.key]

    # def __get_reachable_quests(self):
    #     # cache needs to be reseted before each call
    #     cache: dict[str, bool] = {}
    #     return [q for q in self.quests.values() if Game.__is_reachable_quest(q, cache)]

    def update_reachable_and_available(self):
        for q in self.quests.values():
            q.set_reachable(False)
            q.update_tasks_reachable()
        for c in self.clusters.values():
            c.set_reachable(False)

        cache: dict[str, bool] = {}
        for c in self.clusters.values():
            for q in c.get_quests():
                if Game.is_reachable_quest(q, cache):
                    q.set_reachable(True)
                    c.set_reachable(True)

    # @override
    def __str__(self):
        output: list[str] = []
        for c in self.clusters.values():
            output.append("# " + str(c))
            for q in c.get_quests():
                output.append("  - " + str(q))
                for t in q.get_tasks():
                    output.append("    - " + str(t))
        output.append(100 * "-")
        for q in self.quests.values():
            output.append(str(q))
        output.append(100 * "-")
        for t in self.tasks.values():
            output.append(str(t))
        return "\n".join(output)


from tko.util.text import Token

class __Symbols:

    def __init__(self):
        self.set_unicode()
        # self.set_ascii()

    def set_unicode(self):
        self.task_downloaded = Token("▼", "g")
        self.task_to_download = Token("▲")
        self.task_local = Token("▶", "g")
        self.task_to_visit = Token("◉", "b")
        self.opening = Token("=> ")
        self.neutral = Token("»")
        self.success = Token("✓")
        self.failure = Token("✗")
        self.wrong = Token("ω")
        self.compilation = Token("ϲ")
        self.execution = Token("ϵ")
        self.unequal = Token("├")
        self.equalbar = Token("│")
        self.hbar = Token("─")
        self.vbar = Token("│")

        self.whitespace = Token("·") #Token("␣")
        self.newline = Token("↲")

        self.cfill = Token("_")
        self.tab = Token("    ")
        self.arrow_up = Token("↑")

        self.check = Token("✓")
        self.uncheck = Token("✗")

        self.infinity = Token("∞")
        self.locked_free = Token("⇉")
        self.locked_locked = Token("⇟")
        self.left_toggle = Token("━─")
        self.right_toggle = Token("─━")
        self.timer = Token("⏳") #⏰
        self.diff_down = Token("↓")
        self.diff_left = Token("→")
        self.sharpR = Token("")
        self.sharpL = Token("")
        self.action = Token("◎", "b")

        # self.emoji_confiante = Token("●", "g")
        # self.emoji_capaz     = Token("◕", "y")
        # self.emoji_inseguro  = Token("◑", "m")
        # self.emoji_confuso   = Token("◔", "r")
        # self.emoji_nao_fiz   = Token("ⵔ")

        # self.emoji_confiante = Token("■", "g")
        # self.emoji_capaz     = Token("◨", "y")
        # self.emoji_inseguro  = Token("◧", "m")
        # self.emoji_confuso   = Token("◱", "r")
        # self.emoji_nao_fiz   = Token("□")

        self.autonomy_a = Token("▇", "g")
        self.autonomy_b = Token("▆", "y")
        self.autonomy_c = Token("▅", "m")
        self.autonomy_d = Token("▄", "r")
        self.autonomy_e = Token("▂", "c")
        self.autonomy_x = Token("x", "")
        self.autonomy_list = [self.autonomy_x, self.autonomy_e, self.autonomy_d, self.autonomy_c, self.autonomy_b, self.autonomy_a]

        self.approach_s = Token("S", "b")
        self.approach_a = Token("A", "g")
        self.approach_b = Token("B", "y")
        self.approach_c = Token("C", "m")
        self.approach_d = Token("D", "r")
        self.approach_e = Token("E", "c")
        self.approach_x = Token("x", "")
        self.approach_list = [self.approach_x, self.approach_e, self.approach_d, self.approach_c, self.approach_b, self.approach_a, self.approach_s]
        
        self.desire_a = Token("5", "g")
        self.desire_b = Token("4", "y")
        self.desire_c = Token("3", "m")
        self.desire_d = Token("2", "r")
        self.desire_e = Token("1", "c")
        self.desire_x = Token("x", "")
        self.desire_list = [self.desire_x, self.desire_e, self.desire_d, self.desire_c, self.desire_b, self.desire_a]

        self.star = Token("★", "g")
        self.open_star = Token("☆")
        self.mark = Token("✦", "g")

        self.cursor = Token("┊")

    # def set_ascii(self):
    #     self.task_downloaded = Token("D", "g")
    #     self.task_to_download = Token("X")
    #     self.task_local = Token(">", "g")
    #     self.task_to_visit = Token("◉", "b")
    #     self.whitespace = Token("¨")
    #     self.newline = Token("~")

    #     self.cfill = Token("_")
    #     self.tab = Token("    ")
    #     self.arrow_up = Token("|")

    #     self.check = Token("A")
    #     self.uncheck = Token("0")

    #     self.infinity = Token("0")
    #     self.locked_free = Token(">")
    #     self.locked_locked = Token("v")
    #     self.left_toggle = Token("━─")
    #     self.right_toggle = Token("─━")
    #     self.timer = Token("l")
    #     self.diff_down = Token("|")
    #     self.diff_left = Token("─")
    #     self.sharpR = Token("")
    #     self.sharpL = Token("")
    #     self.action = Token("◎", "b")

    #     self.emoji_vazio = Token("✗", "r")
    #     self.emoji_confiante = Token("A", "g")
    #     self.emoji_capaz     = Token("B", "y")
    #     self.emoji_inseguro  = Token("C", "m")
    #     self.emoji_confuso   = Token("D", "r")
    #     self.emoji_nao_fiz   = Token("E")

    #     self.emoji_alone = Token("A", "g")
    #     self.emoji_dicas = Token("B", "y")
    #     self.emoji_codes = Token("C", "m")
    #     self.emoji_guide = Token("D", "r")
    #     self.cursor = Token("|")

    def set_colors(self):
        self.opening.fmt = "b"
        self.neutral.fmt = "b"
        self.success.fmt = "g"
        self.failure.fmt = "r"
        self.wrong.fmt = "r"
        self.compilation.fmt = "y"
        self.execution.fmt = "y"
        self.unequal.fmt = "r"
        self.equalbar.fmt = "g"

symbols = __Symbols()



from tko.settings.logger import Logger
from uniplot import plot_to_string # type: ignore
from tko.util.text import Text

class WeekGraph:
    def __init__(self, width: int, height: int, week_mode: bool = False):
        self.width = width
        self.height = height
        self.week_mode = week_mode
        self.collected: list[float] = []
        self.eixo: list[float] = []
        self.__collect()

    def __collect(self):
        logger = Logger.get_instance()
        days_minutes = logger.week.resume()
        if self.week_mode:
            self.collected = [0]
            for _, value in days_minutes.items():
                week_day = value.weed_day
                elapsed = value.elapsed
                if week_day == "Sunday":
                    self.collected.append(0)
                self.collected[-1] += elapsed

        else:
            self.collected = []
            for value in days_minutes.values():
                self.collected.append(value.elapsed)
        # convertendo de minutos pra horas/minutos
        for i in range(len(self.collected)):
            value = self.collected[i]
            hours = value // 60
            minutes = value % 60
            self.collected[i] = hours + minutes / 100.0
        self.eixo = list(range(len(self.collected)))
        # self.eixo = list(range(len(collected)))

    def get_collected(self) -> list[str]:
        output: list[str] = []
        for value in self.collected:
            output.append(f"{value:.2f}")
        for i in range(len(output)):
            output[i] = output[i].replace(".", ":").rjust(5, " ")
        return output

    def get_graph(self) -> list[Text]:
        collected: list[float] = []
        eixo: list[int] = []
        if self.week_mode:
            title = " (horas / semana)"
            week = 1
            for value in self.collected:
                eixo.append(week)
                collected.append(0)
                eixo.append(week)
                collected.append(value)
                eixo.append(week)
                collected.append(0)
                week += 1
            result = plot_to_string(xs=eixo, ys=collected, title=title, lines=True, y_min=0, width=self.width, height=self.height, y_unit="h", x_unit="s")
        else:
            title = " (horas /  dia  )"
            day = 1
            for value in self.collected:
                eixo.append(day)
                collected.append(0)
                eixo.append(day)
                collected.append(value)
                eixo.append(day)
                collected.append(0)
                day += 1

            result = plot_to_string(xs=eixo, ys=collected, title=title, lines=True, y_min=0, width=self.width, height=self.height, y_unit="h", x_unit="d")
        if isinstance(result, str):
            return [Text().add(x) for x in result.splitlines()]
        return [Text().add(x) for x in result]



from tko.util.text import Text
from tko.play.fmt import Fmt

class TestSimple:
    def test_split_1(self):
        values = Fmt.cut_box(6, 6, 10, 10, Text().add("1234\n56\n789").addf("r", "1\n23").addf("g", "456"))
        tokens = Fmt.split_in_tokens(values)
        assert ", ".join([str(x) for x in tokens]) == "6:6:(:1234), 7:6:(:56), 8:6:(:789), 8:9:(r:1), 9:6:(r:23), 9:8:(g:45)"

    def test_split_2(self):
        values = Fmt.cut_box(-2, -1, 10, 10, Text().add("1234\n56\n789").addf("r", "1\n23").addf("g", "456"))
        tokens = Fmt.split_in_tokens(values)
        assert ", ".join([str(x) for x in tokens]) == "0:0:(:89), 0:2:(r:1), 1:0:(r:3), 1:1:(g:456)"



import yaml # type: ignore
from tko.settings.log_action import LogAction
from tko.util.text import Text

class CacheListener:
    def __init__(self):
        self.cache: dict[str, list[Text]] = {}

    def listener(self, action: LogAction, new_entry: bool = False):
        if action.task_key == "":
            return
        self.cache[action.task_key] = []

    def set_task_cache(self, key: str, lines: list[Text]):
        self.cache[key] = lines

    def get_task_cache(self, key: str) -> list[Text]:
        if key not in self.cache:
            return []
        return self.cache[key]
    


from tko.game.quest import Quest
from tko.util.text import Text
from tko.game.tree_item import TreeItem
# from typing import override

class Cluster(TreeItem):
    def __init__(self, line_number: int = 0, title: str = "", key: str = "", color: None | str = None):
        super().__init__()
        self.line_number = line_number
        self.__quests: list[Quest] = []
        self.color: None | str = color
        self.__is_reachable = False
        self.key = key
        self.title = title

    def add_quest(self, quest: Quest):
        self.__quests.append(quest)
        return self

    def remove_empty_or_other_language(self, language: str):
        # self.__quests = [q for q in self.__quests if len(q.get_tasks()) > 0]
        quests: list[Quest] = []
        for q in self.__quests:
            if len(q.get_tasks()) == 0:
                continue
            if len(q.languages) == 0 or language in q.languages:
                quests.append(q)
        self.__quests = quests

        return self

    def get_quests(self) -> list[Quest]:
        return self.__quests

    def is_reachable(self):
        return self.__is_reachable
    
    def set_reachable(self, value: bool):
        self.__is_reachable = value
        return self

    # @override
    def __str__(self):
        line = str(self.line_number).rjust(3)
        quests_size = str(len(self.__quests)).rjust(2, "0")
        key = "" if self.key == self.title else self.key + " "
        return f"{line} {quests_size} {key}{self.title}"
    
    def get_grade_color(self) -> str:
        perc = self.get_percent()
        if perc == 0:
            return "m"
        if perc < 50:
            return "r"
        if perc < 100:
            return "y"
        return "g"

    def get_percent(self):
        total = 0
        for q in self.__quests:
            total += q.get_percent()
        return total // len(self.__quests)

    def get_resume_by_percent(self) -> Text:
        return Text().addf(self.get_grade_color(), f"{self.get_percent()}%".rjust(4))

    def get_resume_by_quests(self):
        total = len(self.__quests)
        count = len([q for q in self.__quests if q.is_complete()])
        return Text().addf(self.get_grade_color(), f"({count}/{total})")
        



from tko.game.game import Game

class XP:
    token_level_one = "level_one"
    token_level_mult = "level_mult"
    level_one: int = 100
    level_mult: float = 1.5
    
    def __init__(self, game: Game):
        self.game = game
        self.obtained = 0
        self.avaliable = 0
        self.update()
        self.level = self.get_level()
    
    def update(self):
        self.obtained, self.avaliable = self.game.get_xp_resume()

    def get_level(self) -> int:
        return self.calc_level(self.obtained)

    def get_xp_level_current(self) -> int:
        xp_prev = self.calc_xp(self.level)
        atual = self.obtained - xp_prev
        return atual

    def get_xp_level_needed(self) -> int:
        xp_next = self.calc_xp(self.level + 1)
        xp_prev = self.calc_xp(self.level)
        return xp_next - xp_prev

    def get_xp_total_obtained(self) -> int:
        return self.obtained

    def get_xp_total_available(self) -> int:
        return self.avaliable

    def calc_level(self, xp: int) -> int:
        level = 1
        while self.calc_xp(level) <= xp:
            level += 1
        return level - 1
    
    def calc_xp(self, level: int) -> int:
        total = 0
        for i in range(level - 1):
            total += self.game.level_one * (int(self.game.level_mult) ** i)
        return int(total)



def get_md_link(title: str) -> str:
    title = title.lower()
    out = ""
    for c in title:
        if c == " " or c == "-":
            out += "-"
        elif c == "_":
            out += "_"
        elif c.isalnum():
            out += c
    return out


from tko.play.flags import Flag
from typing import Callable

class FlagFunctor:
    def __init__(self, flag: Flag):
        self.flag = flag

    def __call__(self):
        self.flag.toggle()

class GradeFunctor:
    def __init__(self, grade: int, fn: Callable[[int], None]):
        self.grade: int = grade
        self.fn = fn

    def __call__(self):
        self.fn(self.grade)



import os
from tko.util.text import Text
from tko.util.decoder import Decoder
from tko.settings.repository import Repository
from tko.settings.settings import Settings

class RepStarter:
    def __init__(self, remote: str | None, source: str | None, folder: str | None, language: str | None = None, database: str | None = None):
        self.folder: str = ""
        # if folder is set, use folder, else use remote if remote, else .
        if not self.set_folder(folder, remote):
            return

        rep = self.createRepository()
        if rep is None:
            return
        self.rep = rep

        if source is None and remote is None:
            print(Text.format("Nenhuma fonte foi informada, criando repositório vazio."))
            self.create_empty_rep()
        elif remote is not None:
            print(Text.format("Criando repositório apontando para repositório remoto {y}.", remote))
            self.source: str = ""
            settings = Settings()
            if not settings.has_alias_remote(remote):
                raise Warning("fail: alias remoto não encontrado.")
            self.source = settings.get_alias_remote(remote)
            rep.set_rep_source(self.source)
        elif source is not None:
            if source.startswith("http"):
                print(Text.format("Criando repositório apontando para link remota {y}.", source))
                self.source = source
            else:
                print(Text.format("Criando repositório apontando para arquivo local {y}.", source))
                self.source = os.path.abspath(source)
                if self.source.startswith(self.folder):
                    self.source = os.path.relpath(self.source, os.path.dirname(rep.get_config_file()))
            rep.set_rep_source(self.source)
        if language is not None:
            rep.set_lang(language)
            print(Text.format("A linguagem do repositório foi definida como {y}.", language))
        if database is not None:
            rep.set_database_folder(database)
            print(Text.format("A pasta com as questões foi definida para {y}.", database))
        rep.save_config()
        self.print_end_msg()

    def print_end_msg(self):
        rel_path = os.path.relpath(self.rep.root_folder, os.getcwd())
        print(Text.format("Voce pode acessar o repositório com o comando {g} {y}", "tko open", "<pasta>"))
        print(Text.format("Por exemplo: {g} {y}", "tko open", rel_path))

    def set_folder(self, folder: str | None, remote: str | None) -> bool:
        if folder is not None:
            self.folder = os.path.abspath(folder)
            return True
        
        self.folder = os.path.abspath(os.getcwd())
        if remote is not None:
            self.folder = os.path.join(self.folder, remote)
        print(Text.format("A pasta onde deve ser criada o repositório {r:não} foi informada."))
        print(Text.format("Deseja criar o repositório na pasta {y} ? ({g}/{r}): ", self.folder, "s", "n"), end="")
        op = input()
        if op == "n":
            return False
        return True
    
    def createRepository(self) -> Repository | None:
        path_old = Repository.rec_search_for_repo(self.folder)
        if path_old != "":
            if self.folder != path_old:
                print(Text.format("Você está tentando criar um rep dentro de outro, pois já existe rep em {r}.", path_old))
                print(Text.format("Você pode apagar o repositório antigo removendo a pasta {r:}.", os.path.join(path_old, ".tko")))
                print(Text.format("Ou sobrescrever as configurações de  {r:}.", os.path.join(path_old, ".tko")))
            self.folder = path_old
            print(Text.format("Deseja sobrescrever as configurações do repositório em {y} ? ({g}/{r}): ", self.folder, "s", "n"), end="")
            op = input()
            if op == "n":
                return None
        return Repository(os.path.abspath(self.folder))
    
    def create_empty_rep(self):
        index = self.rep.get_default_readme_path()
        self.rep.set_rep_source(index)
        print("Nenhuma fonte foi informada, utilizando o arquivo {} como fonte".format(index))
        if not os.path.exists(index):
            content = "# Repositório\n\n## Grupo\n\n### Missão\n\n- [ ] [#google Abra o google](https://www.google.com)\n"
            Decoder.save(index, content)


from tko.play.floating import FloatingInputData, FloatingInput
from tko.util.text import Text, Token
from tko.util.symbols import symbols
from tko.play.keys import GuiKeys
from tko.play.flags import Flags
from tko.play.play_actions import PlayActions
from tko.play.floating_calibrate import FloatingCalibrate

class PlayConfig:
    def __init__(self, actions: PlayActions):
        self.actions = actions
        self.app = self.actions.settings.app
        self.fman = self.actions.fman
        self.gui = self.actions.gui

    def command_pallete(self):
        options: list[FloatingInputData] = []

        def icon(value: bool):
            return Token("✓", "g") if value else Token("✗", "r")
        
        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Tarefa: {y} para repositório local", symbols.action, "Baixar"),
                self.actions.down_remote_task,
                GuiKeys.down_task
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Tarefa: Abrir {y} com a descrição", symbols.action, "URL"),
                self.actions.open_link,
                GuiKeys.open_url
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Tarefa: {y} arquivos na IDE", symbols.action, "Editar"),
                self.actions.open_code,
                GuiKeys.edit
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Muda a {y} do gráfico", icon(Flags.graph.is_true()), "Visão"),
                Flags.graph.toggle,
                Flags.graph.get_keycode()
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y}", symbols.action, "Ajuda"),
                self.gui.show_help,
                GuiKeys.key_help
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y}", icon(self.app.get_use_borders()), "Bordas"),
                self.app.toggle_borders,
                GuiKeys.borders
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y}", icon(self.app.get_use_images()), "Imagens"),
                self.app.toggle_images, 
                GuiKeys.images
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar atividades {y}", icon(self.app.get_show_hidden()), "Escondidas"),
                self.app.toggle_hidden, 
                GuiKeys.hidden
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y} ou Estrelas", icon(Flags.percent.is_true()), "Percentual"),
                Flags.percent.toggle, 
                Flags.percent.get_keycode()
            )
        )

        # options.append(
        #     FloatingInputData(
        #         lambda: Text.format(" {} Mostrar {y} para completar a missão", icon(Flags.minimum.is_true()), "Mínimo"),
        #         Flags.minimum.toggle,
        #         Flags.minimum.get_keycode()
        #     )
        # )

        # options.append(
        #     FloatingInputData(
        #         lambda: Text.format(" {} Mostrar {y} das tarefas", icon(Flags.reward.is_true()), "Ganho"),
        #         Flags.reward.toggle, 
        #         Flags.reward.get_keycode()
        #     )
        # )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mostrar {y} adquiridas", icon(Flags.skills.is_true()), Flags.skills.get_name()),
                Flags.skills.toggle, 
                Flags.skills.get_keycode()
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Modo {y}: Habilita apenas tarefas recomendadas", icon(not Flags.admin.is_true()), "Gamer"),
                Flags.admin.toggle,
                Flags.admin.get_keycode()
            )
        )

        # options.append(
        #     FloatingInputData(
        #         lambda: Text.format(" {} Gerar {y} de dependências", symbols.action, "Grafo"),
        #         self.actions.generate_graph,
        #         GuiKeys.graph
        #     ).set_exit_on_action(True)
        # )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} Mudar linguagem de download de {y}", symbols.action, "Rascunhos"),
                self.gui.language.set_language,
                GuiKeys.set_lang_drafts
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text.format(" {} {y} as teclas direcionais", symbols.action, "Calibrar"),
                lambda: self.fman.add_input(FloatingCalibrate(self.actions.settings)),
                GuiKeys.calibrate
            ).set_exit_on_action(True)
        )

        self.fman.add_input(
            FloatingInput(self.actions.settings, "^").set_text_ljust()
                      .set_header(" Selecione uma ação da lista ")
                      .set_options(options)
                      .set_exit_on_enter(False)
                      .set_footer(" Use Enter para aplicar e Esc para Sair ")
        )


import tempfile

import os
import shutil
from tko.util.text import Text
from tko.util.runner import Runner
from tko.util.decoder import Decoder
import yaml #type: ignore
import io

# from typing import override

class CompileError(Exception):
    def __init__(self, message: str):
        self.message = message

    # @override
    def __str__(self):
        return self.message

class Executable:
    def __init__(self, cmd: list[str] = [], files: list[str] = [], folder: str | None= None):
        self.__cmd = cmd
        self.__files = files
        self.__folder = folder
        self.__compiled: bool = False
        self.__compile_error = False
        self.__error_msg: Text = Text()
    
    def set_executable(self, cmd: list[str], files: list[str], folder: str | None = None):
        self.__compiled = True
        self.__cmd = cmd
        if folder is not None and " " in folder:
            folder = '"' + folder + '"'
        self.__files = fix_spaces(files)
        self.__folder = folder
        return self
    
    def get_command(self) -> tuple[str, str | None]:
        cmd = " ".join(self.__cmd) + " " + " ".join(self.__files)
        return cmd, None if self.__folder is None else self.__folder

    def set_compile_error(self, error_msg: Text | str):
        self.__compiled = True
        self.__compile_error = True
        if isinstance(error_msg, str):
            self.__error_msg = Text().add(error_msg)
        else:
            self.__error_msg = error_msg
        return self
    
    def compiled(self):
        return self.__compiled

    def has_compile_error(self):
        return self.__compiled and self.__compile_error
    
    def get_error_msg(self):
        return self.__error_msg

def fix_spaces(file_list: list[str]) -> list[str]:
    fixed_list: list[str] = []
    for path in file_list:
        if " " in path and not path.startswith('"'):
            fixed_list.append(f'"{path}"')
        else:
            fixed_list.append(path)
    return fixed_list

class SolverBuilder:
    def __init__(self, args_list: list[str]):
        self.args_list: list[str] = [os.path.normpath(SolverBuilder.__add_dot_bar(path)) for path in args_list]

        if len(self.args_list) > 0:
            self.cache_dir = os.path.join(os.path.dirname(self.args_list[0]), ".build")
        else:
            self.cache_dir = os.path.abspath(tempfile.mkdtemp())
        self.clear_cache()
        self.__exec: Executable = Executable()

    def check_tool(self, name: str):
        if shutil.which(name) is None:
            self.__exec.set_compile_error(Text.format("{r}: O comando '" + name + "' não foi encontrado", "Falha")) 
            raise CompileError("fail: comando '" + name + "' não foi encontrado")

    def not_compiled(self):
        return not self.__exec.compiled()

    def has_compile_error(self):
        return self.__exec.has_compile_error()

    def set_main(self, main: str):
        list_main: list[str] = []
        list_other: list[str] = []

        for path in self.args_list:
            if os.path.basename(path) == main:
                list_main.append(path)
            else:
                list_other.append(path)

        self.args_list = list_main + list_other
        return self

    def clear_cache(self):
        if os.path.exists(self.cache_dir):
            shutil.rmtree(self.cache_dir, ignore_errors=True)
        os.makedirs(self.cache_dir, exist_ok=True)

    def get_executable(self, force_rebuild: bool = False) -> tuple[Executable, bool]:
        if self.args_list == []:
            return Executable(), False
        if self.__exec.has_compile_error():
            return self.__exec, False
        if not self.__exec.compiled() or force_rebuild:
            self.prepare_exec()
        return self.__exec, True

    def reset(self):
        self.__exec = Executable()
        self.clear_cache()

    def prepare_exec(self, free_run_mode: bool = False) -> None:
        self.reset()
        first = self.args_list[0]

        if first.endswith(".py"):
            self.__exec.set_executable(["python3"], self.args_list, "")
        elif first.endswith(".yaml"):
            self.__prepare_yaml()
        elif first.endswith(".mk"):
            self.__prepare_make()
        elif first.endswith(".js"):
            self.__prepare_js(free_run_mode)
        elif first.endswith(".ts"):
            self.__prepare_ts(free_run_mode)
        elif first.endswith(".java"):
            self.__prepare_java()
        elif first.endswith(".hs"):
            self.__prepare_hs()
        elif first.endswith(".c") or first.endswith(".h"):
            self.__prepare_c()
        elif first.endswith(".cpp") or first.endswith(".hpp"):
            self.__prepare_cpp()
        elif first.endswith(".go"):
            self.__prepare_go()
        else:
            self.__exec.set_executable(self.args_list, [], "")

    def __prepare_java(self):
        self.check_tool("javac")

        first = self.args_list[0]

        filename = os.path.basename(first)
        cmd = ["javac"] + fix_spaces(self.args_list) + ['-d', self.cache_dir]
        return_code, stdout, stderr = Runner.subprocess_run(" ".join(cmd))
        if return_code != 0:
            self.__exec.set_compile_error(stdout + stderr)
        else:
            self.__exec.set_executable(["java", "-cp "], [self.cache_dir, filename[:-5]])  # removing the .java

    def update_input_function(self, free_run_mode: bool, path_list: list[str], copy_dir: str):
        new_files: list[str] = []
        for origin in self.args_list:
            new_files.append(os.path.join(copy_dir, os.path.basename(origin)))

        for i in range(len(path_list)):
            origin = path_list[i]
            destiny = new_files[i]

            encoding = Decoder.get_encoding(origin)
            with open(origin, "r", encoding=encoding) as f:
                content = f.read()
                lines = content.splitlines(keepends=True)

            io_source = io.StringIO()
            io_target = io.StringIO()

            tag = "function input()"
            tag2 = "constinput=()=>"

            if free_run_mode:
                input_cmd = r'function input(): string { let X: any = input; X.P = X.P || require("readline-sync"); return X.P.question(); }'
                input_tag = 'const input = () => ""; // INTERATIVO\n'
            else:
                input_cmd = r'function input(): string { let X: any = input; X.L = X.L || require("fs").readFileSync(0).toString().split(/\r?\n/); return X.L.shift(); } '
                input_tag = 'const input = () => ""; // MODO_TESTE\n'

            inserted: bool = False
            for line in lines:
                match = False
                if not inserted:
                    filtered = "".join([c for c in line if c != " "])
                    match = filtered.startswith(tag2) or line.startswith(tag)
                if match:
                    inserted = True
                    io_source.write(input_tag)
                    io_target.write(input_cmd)
                else:
                    io_source.write(line)
                    io_target.write(line)

            result = io_source.getvalue()
            if result != content:
                with open(origin, "w", encoding="utf-8") as f:
                    f.write(io_source.getvalue())
            with open(destiny, "w", encoding="utf-8") as f:
                f.write(io_target.getvalue())
            io_source.close()
            io_target.close()
        return new_files

    def __prepare_yaml(self):
        solver = os.path.abspath(self.args_list[0])
        folder = os.path.dirname(solver)
        content = Decoder.load(solver)
        yaml_data = yaml.safe_load(content)

        if "build" in yaml_data and yaml_data["build"] is not None:
            build_txt = yaml_data["build"]
            if not isinstance(build_txt, str):
                raise Warning(Text.format("{r}: O campo build deve ser uma string", "Falha"))
            
            return_code, stdout, stderr = Runner.subprocess_run(build_txt, folder = folder)
            if return_code != 0:
                self.__exec.set_compile_error(stdout + stderr)
                return
        if not "run" in yaml_data:
            raise Warning(Text.format("{r}: Seu arquivo yaml precisa ter um campo {g} ", "Falha", "run:"))
        if not isinstance(yaml_data["run"], str):
            raise Warning(Text.format("{r}: O campo run deve ser uma lista", "Falha"))
        self.__exec.set_executable([yaml_data["run"]], [], folder)

    def __prepare_make(self):
        self.check_tool("make")
        solver = os.path.abspath(self.args_list[0])
        folder = os.path.dirname(solver)
        cmd = ["make", "-s", "-C", folder, "-f", solver, "build"]
        return_code, stdout, stderr = Runner.subprocess_run(" ".join(cmd))
        if return_code != 0:
            self.__exec.set_compile_error(stdout + stderr)
        else:
            self.__exec.set_executable(["make", "-s", "-C", folder, "-f", solver, "run"], [])

    def __prepare_js(self, free_run_mode: bool):
        self.check_tool("node")
        self.__exec.set_executable(["node"], self.args_list)

    def __prepare_go(self):
        self.check_tool("go")
        
        self.__exec.set_executable(["go", "run"], self.args_list)

    def __prepare_ts(self, free_run_mode: bool):
        copy_dir = os.path.join(self.cache_dir, "src")
        # remove the cache dir
        if os.path.exists(copy_dir):
            shutil.rmtree(copy_dir, ignore_errors=True)
        os.makedirs(copy_dir, exist_ok=True)
        new_files = self.update_input_function(free_run_mode, self.args_list, copy_dir)
        transpiler = "npx"
        if os.name == "nt":
            transpiler += ".cmd"

        new_files = [os.path.abspath(x) for x in new_files]
        self.cache_dir = os.path.abspath(self.cache_dir)

        self.check_tool(transpiler)
        self.check_tool("node")
        transpiler = "npx esbuild"
        cmd = [transpiler] + new_files + ["--outdir=" + self.cache_dir, "--format=cjs", "--log-level=error"]
        return_code, stdout, stderr = Runner.subprocess_run(" ".join(cmd))

        if return_code != 0:
            self.__exec.set_compile_error(stdout + stderr)
        else:
            new_source_list: list[str] = []
            for source in new_files:
                new_source_list.append(os.path.join(self.cache_dir, os.path.basename(source)[:-2] + "js"))
            self.__exec.set_executable(["node"], new_source_list)  # renaming solver to main

    def __prepare_c_cpp(self, pre_args: list[str], pos_args: list[str]):
        # solver = self.path_list[0]
        tempdir = self.cache_dir
        source_list = [x for x in self.args_list if not x.endswith(".h") and not x.endswith(".hpp")]
        exec_path = os.path.join(tempdir, ".a.out")
        cmd = pre_args + source_list + ["-o", exec_path] + pos_args
        return_code, stdout, stderr = Runner.subprocess_run(" ".join(cmd))
        if return_code != 0:
            self.__exec.set_compile_error(stdout + stderr)
        else:
            self.__exec.set_executable([], [exec_path])

    def __prepare_hs(self):
        self.check_tool("runhaskell")
        self.check_tool("ghc")
        self.__exec.set_executable(["runhaskell"], self.args_list, "")


    def __prepare_c(self):
        self.check_tool("gcc")
        pre = ["gcc", "-Wall"]
        pos = ["-lm"]
        self.__prepare_c_cpp(pre, pos)

    def __prepare_cpp(self):
        self.check_tool("g++")
        pre = ["g++", "-std=c++17", "-Wall", "-Wextra", "-Werror", "-Wno-deprecated"]
        pos: list[str] = []
        self.__prepare_c_cpp(pre, pos)

    @staticmethod
    def __add_dot_bar(solver: str) -> str:
        if os.sep not in solver and os.path.isfile("." + os.sep + solver):
            solver = "." + os.sep + solver
        return solver
